# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Utilities for synchronization."""

import os
import shutil
import filecmp
import logging
from copy import deepcopy
from contextlib import contextmanager
from filecmp import dircmp


LEVEL_MORE = logging.INFO - 5

logger = logging.getLogger('sync')
logging.addLevelName(LEVEL_MORE, 'MORE')
logging.MORE = LEVEL_MORE  # type: ignore


def log_more(msg, *args, **kwargs):
    """Log using LEVEL_MORE."""
    logger.log(LEVEL_MORE, msg, *args, **kwargs)


logger.more = log_more  # type: ignore


def copytree(src, dst, copy_function=shutil.copy2, symlinks=False):
    """Recursively copy a directory tree from src to dst, using a custom copy function.

    Implementation adapted from https://docs.python.org/3/library/shutil.html#copytree-example.

    Parameters
    ----------
    src : str
        Source directory tree.
    dst : str
        Destination directory tree.
    copy_function :
        Function used to copy (Default value = ``shutil.copy2``).
    symlinks : bool
        Whether to copy symlinks (Default value = False).

    """
    os.makedirs(dst)
    names = os.listdir(src)
    errors = []
    for name in names:
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        try:
            if symlinks and os.path.islink(srcname):
                linkto = os.readlink(srcname)
                os.symlink(linkto, dstname)
            elif os.path.isdir(srcname):
                copytree(srcname, dstname, copy_function, symlinks)
            else:
                copy_function(srcname, dstname)
        except OSError as why:
            errors.append((srcname, dstname, str(why)))
        # catch the Error from the recursive copytree so that we can
        # continue with other files
        except shutil.Error as err:
            errors.extend(err.args[0])
    if errors:
        raise shutil.Error(errors)


class dircmp_deep(dircmp):
    """Deep directory comparator."""
    def phase3(self):
        """Find out differences between common files."""
        xx = filecmp.cmpfiles(self.left, self.right, self.common_files, shallow=False)
        self.same_files, self.diff_files, self.funny_files = xx

    methodmap = dict(dircmp.methodmap)
    # The type check for the following line must be ignored.
    # See: https://github.com/python/mypy/issues/708
    methodmap['same_files'] = methodmap['diff_files'] = phase3  # type: ignore


class _DocProxy(object):
    """Proxy object for document (mapping) modifications.

    This proxy is used to keep track of changes and ensure that
    dry runs do not actually modify any data.

    Parameters
    ----------
    doc : dict
        Document data.
    dry_run : bool
        Do not actually perform any data modification operation, but still log
        the action (Default value = False).

    """

    def __init__(self, doc, dry_run=False):
        self.doc = doc
        self.dry_run = dry_run

    def __str__(self):
        return "_DocProxy({})".format(str(self.doc))

    def __repr__(self):
        return "_DocProxy({})".format(repr(self.doc))

    def __getitem__(self, key):
        return self.doc[key]

    def __setitem__(self, key, value):
        logger.more("Set '{}'='{}'.".format(key, value))
        if not self.dry_run:
            self.doc[key] = value

    def keys(self):
        """Return keys of proxy data."""
        return self.doc.keys()

    def clear(self):
        """Clear proxy data."""
        self.doc.clear()

    def update(self, other):
        """Update proxy data with other."""
        for key in other.keys():
            self[key] = other[key]

    def __iter__(self):
        return iter(self.doc)

    def __contains__(self, key):
        return key in self.doc

    def __eq__(self, other):
        return self.doc.__eq__(other)

    def __len__(self):
        return len(self.doc)


class _FileModifyProxy(object):
    """Proxy used for data modification.

    This proxy is used for all file data modification to keep
    track of changes and to ensure that dry runs do not actually
    modify any data.

    Parameters
    ----------
    root :
        Root path.
    follow_symlinks : bool
        Whether to follow symlinks (Default value = True).
    permissions : bool
        Whether to preserve permissions (Default value = False).
    times : bool
        Whether to preserve timestamps (Default value = False).
    owner : bool
        Whether to preserve owner (Default value = False).
    group : bool
        Whether to preserve group (Default value = False).
    dry_run : bool
        If True, do not actually perform any data modification operation, but still log
        the action (Default value = False).
    collect_stats : bool
        Whether to collect stats (Default value = False).

    """

    def __init__(self, root=None, follow_symlinks=True, permissions=False,
                 times=False, owner=False, group=False, dry_run=False,
                 collect_stats=False):
        self.root = root
        self.follow_symlinks = follow_symlinks
        self.permissions = permissions
        self.times = times
        self.owner = owner
        self.group = group
        self.dry_run = dry_run
        self.stats = dict(num_files=0, volume=0) if collect_stats else None

    # Internal proxy functions

    def _copy(self, src, dst):
        """Copy src to dst."""
        if not self.dry_run:
            shutil.copy(src, dst)

    def _copy_p(self, src, dst):
        """Copy src to dst with permissions."""
        if not self.dry_run:
            shutil.copy(src, dst)
            shutil.copymode(src, dst)

    def _copy2(self, src, dst):
        """Copy src to dst with preserved metadata."""
        if not self.dry_run:
            shutil.copy2(src, dst)

    def _remove(self, path):
        """Remove path."""
        if not self.dry_run:
            os.remove(path)

    # Public functions

    def remove(self, path):
        """Remove path."""
        logger.more("Remove path '{}'.".format(os.path.relpath(path)))
        self._remove(path)

    def copy(self, src, dst):
        """Copy src to dst."""
        if self.dry_run and self.root is not None:
            print(os.path.relpath(src, self.root))
        if os.path.islink(src) and not self.follow_symlinks:
            link_target = os.readlink(src)
            logger.more("Creating link '{}' -> '{}'.".format(
                os.path.relpath(dst), os.path.relpath(link_target)))
            if os.path.isfile(dst):
                self.remove(dst)
            if not self.dry_run:
                os.symlink(link_target, dst)
        else:
            msg = "Copy file{{}} '{}' -> '{}'.".format(os.path.relpath(src), os.path.relpath(dst))
            if self.permissions and self.times:
                logger.more(msg.format(' (preserving: permissions, times)'))
                self._copy2(src, dst)
            elif self.permissions:
                logger.more(msg.format(' (preserving: permissions)'))
                self._copy_p(src, dst)
            elif self.times:
                raise ValueError("Cannot copy timestamps without permissions.")
            else:
                logger.more(msg.format(''))
                self._copy(src, dst)
            if self.owner or self.group or self.stats is not None:
                stat = os.stat(src)
                if self.stats is not None:
                    self.stats['num_files'] += 1
                    self.stats['volume'] += stat.st_size
                if self.owner or self.group:
                    logger.more("Copy owner/group '{}' -> '{}'".format(
                        os.path.relpath(src), os.path.relpath(dst)))
                    if not self.dry_run:
                        os.chown(dst,
                                 uid=stat.st_uid if self.owner else -1,
                                 gid=stat.st_gid if self.group else -1)

    def copytree(self, src, dst, **kwargs):
        """Copy tree src to dst."""
        logger.more("Copy tree '{}' -> '{}'.".format(os.path.relpath(src), os.path.relpath(dst)))
        copytree(src, dst, copy_function=self.copy, **kwargs)

    @contextmanager
    def create_backup(self, path):
        """Create a backup of path."""
        logger.debug("Create backup of '{}'.".format(os.path.relpath(path)))
        path_backup = path + '~'
        if os.path.isfile(path_backup):
            raise RuntimeError(
                "Failed to create backup, file already exists: '{}'.".format(
                    os.path.relpath(path_backup)))
        try:
            self._copy2(path, path_backup)
            yield path_backup
        except:     # noqa roll-back
            logger.more("Error occured, restoring backup...")
            self._copy2(path_backup, path)
            raise
        finally:
            logger.debug("Remove backup of '{}'.".format(os.path.relpath(path)))
            self._remove(path_backup)

    @contextmanager
    def create_doc_backup(self, doc):
        """Create a backup of doc."""
        proxy = _DocProxy(doc, dry_run=self.dry_run)
        fn = getattr(doc, 'filename', getattr(doc, '_filename', None))
        if not len(proxy) or fn is None or not os.path.isfile(fn):
            backup = deepcopy(doc)  # use in-memory backup
            try:
                yield proxy
            except:     # noqa roll-back
                proxy.clear()
                proxy.update(backup)
                raise
        else:
            with self.create_backup(fn):
                yield proxy



# Copyright (c) 2019 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Handle migrations of signac schema versions."""

import os
import sys
from packaging import version
from contextlib import contextmanager

from filelock import FileLock

from ...common.config import get_config
from ...version import __version__, SCHEMA_VERSION


from .v0_to_v1 import migrate_v0_to_v1


FN_MIGRATION_LOCKFILE = '.SIGNAC_PROJECT_MIGRATION_LOCK'


MIGRATIONS = {
    ('0', '1'):    migrate_v0_to_v1,
}


def _reload_project_config(project):
    project_reloaded = project.get_project(
        root=project._rd, search=False, _ignore_schema_version=True)
    project._config = project_reloaded._config


def _update_project_config(project, **kwargs):
    """Update the project configuration."""
    for fn in ('signac.rc', '.signacrc'):
        config = get_config(project.fn(fn))
        if 'project' in config:
            break
    else:
        raise RuntimeError("Unable to determine project configuration file.")
    config.update(kwargs)
    config.write()
    _reload_project_config(project)


@contextmanager
def _lock_for_migration(project):
    lock = FileLock(project.fn(FN_MIGRATION_LOCKFILE))
    try:
        with lock:
            yield
    finally:
        try:
            os.unlink(lock.lock_file)
        except FileNotFoundError:
            pass


def _collect_migrations(project):
    schema_version = version.parse(SCHEMA_VERSION)

    def config_schema_version():
        return version.parse(project._config['schema_version'])

    if config_schema_version() > schema_version:
        # Project config schema version is newer and therefore not supported.
        raise RuntimeError(
            "The signac schema version used by this project is {}, but signac {} "
            "only supports up to schema version {}. Try updating signac.".format(
                config_schema_version, __version__, SCHEMA_VERSION))

    while config_schema_version() < schema_version:
        for (origin, destination), migration in MIGRATIONS.items():
            if version.parse(origin) == config_schema_version():
                yield (origin, destination), migration
                break
        else:
            raise RuntimeError(
                "The signac schema version used by this project is {}, but signac {} "
                "uses schema version {} and does not know how to migrate.".format(
                    config_schema_version(), __version__, schema_version))


def apply_migrations(project):
    """Apply migrations to a project."""
    with _lock_for_migration(project):
        for (origin, destination), migrate in _collect_migrations(project):
            try:
                print("Applying migration for "
                      "version {} to {}... ".format(origin, destination), end='',
                      file=sys.stderr)
                migrate(project)
            except Exception as e:
                raise RuntimeError(
                    "Failed to apply migration {}.".format(destination)) from e
            else:
                _update_project_config(project, schema_version=destination)
                print("OK", file=sys.stderr)
                yield origin, destination


__all__ = [
    'apply_migrations',
    ]



# Copyright (c) 2019 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Compute diffs of state points."""

from .contrib.utility import _dotted_dict_to_nested_dicts
from .contrib.utility import _nested_dicts_to_dotted_keys


def diff_jobs(*jobs):
    r"""Find differences among a list of jobs' state points.

    The resulting diff is a dictionary where the keys are job ids and the
    values are each job's state point minus the intersection of all provided
    jobs' state points. The comparison is performed over the combined set of
    keys and values.

    Parameters
    ----------
    \*jobs : sequence[:class:`~signac.contrib.job.Job`]
        Sequence of jobs to diff.

    Returns
    -------
    dict
        A dictionary where the keys are job ids and values are the unique parts
        of that job's state point.

    Examples
    --------
    >>> import signac
    >>> project = signac.init_project('project_name')
    >>> job1 = project.open_job({'constant': 42, 'diff1': 0, 'diff2': 1}).init()
    >>> job2 = project.open_job({'constant': 42, 'diff1': 1, 'diff2': 1}).init()
    >>> job3 = project.open_job({'constant': 42, 'diff1': 2, 'diff2': 2}).init()
    >>> print(job1)
    c4af2b26f1fd256d70799ad3ce3bdad0
    >>> print(job2)
    b96b21fada698f8934d58359c72755c0
    >>> print(job3)
    e4289419d2b0e57e4852d44a09f167c0
    >>> signac.diff_jobs(job1, job2, job3)
    {'c4af2b26f1fd256d70799ad3ce3bdad0': {'diff2': 1, 'diff1': 0},
    'b96b21fada698f8934d58359c72755c0': {'diff2': 1, 'diff1': 1},
    'e4289419d2b0e57e4852d44a09f167c0': {'diff2': 2, 'diff1': 2}}
    >>> signac.diff_jobs(*project)
    {'c4af2b26f1fd256d70799ad3ce3bdad0': {'diff2': 1, 'diff1': 0},
    'b96b21fada698f8934d58359c72755c0': {'diff2': 1, 'diff1': 1},
    'e4289419d2b0e57e4852d44a09f167c0': {'diff2': 2, 'diff1': 2}}

    """
    if len(jobs) == 0:
        return {}
    else:
        sps = {}
        for job in jobs:
            sps[job] = set(_nested_dicts_to_dotted_keys(job.sp()))

        intersection = set.intersection(*sps.values())

        diffs = {}
        for job in jobs:
            unique_sps = sps[job]-intersection
            diffs[job.id] = _dotted_dict_to_nested_dicts(dict(unique_sps))

        return diffs



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Data store implementation with backend HDF5 file."""
import logging
import os
import errno
import warnings
import array
from threading import RLock
from collections.abc import Mapping
from collections.abc import MutableMapping

from ..errors import InvalidKeyError
from .dict_manager import DictManager


__all__ = [
    'H5Store', 'H5Group', 'H5StoreManager',
    'H5StoreClosedError', 'H5StoreAlreadyOpenError',
    ]


def _group_is_pandas_type(group):
    return 'pandas_type' in group.attrs


_is_pandas_type = None
_pandas = None


def _load_pandas():
    # Late binding to improve import performance.
    global _pandas
    global _is_pandas_type

    if _pandas is None or _is_pandas_type is None:
        try:
            import pandas

            def _is_pandas_type(value):
                return isinstance(value, pandas.core.generic.PandasObject)
        except ImportError:
            def _is_pandas_type(value):
                return False    # Must be False when pandas is not available.
        else:
            _pandas = pandas


def _requires_tables():
    try:
        import tables  # noqa
    except ImportError:
        raise ImportError(
            "Storing and loading pandas objects requires the PyTables package.")


logger = logging.getLogger(__name__)


class H5StoreClosedError(RuntimeError):
    """Raised when trying to access a closed store."""


class H5StoreAlreadyOpenError(OSError):
    """Indicates that the underlying HDF5 file is already openend."""


def _h5set(store, grp, key, value, path=None):
    """Set a key in an h5py container.

    This method recursively converts Mappings to h5py groups and transparently
    handles None values.
    """
    import h5py
    import numpy    # h5py depends on numpy, so this is safe.
    path = path + '/' + key if path else key

    # Guard against assigning a group to itself, e.g., `h5s[key] = h5s[key]`,
    # where h5s[key] is a mapping. This is necessary, because the original
    # mapping would be deleted prior to assignment.
    if key in grp:
        if isinstance(value, H5Group):
            if grp[key] == value._group:
                return  # Groups are identical, do nothing.
        elif isinstance(value, h5py._hl.dataset.Dataset):
            if grp == value.parent:
                return  # Dataset is identical, do nothing.

        # Delete any existing data
        del grp[key]

    # Mapping-types
    if isinstance(value, Mapping):
        subgrp = grp.create_group(key)
        for k, v in value.items():
            _h5set(store, subgrp, k, v, path)

    # Regular built-in types:
    elif value is None:
        grp.create_dataset(key, data=None, shape=None, dtype='f')
    elif isinstance(value, (int, float, str, bool, array.array)):
        grp[key] = value
    elif isinstance(value, bytes):
        grp[key] = numpy.bytes_(value)

    # NumPy types
    elif type(value).__module__ == numpy.__name__:
        grp[key] = value

    # h5py native types
    elif isinstance(value, h5py._hl.dataset.Dataset):
        grp[key] = value  # Creates hard-link!

    # Other types
    else:
        _load_pandas()   # might be a pandas type
        if _is_pandas_type(value):
            _requires_tables()
            store.close()
            with _pandas.HDFStore(store._filename, mode='a') as store_:
                store_[path] = value
            store.open()
        else:
            grp[key] = value
            warnings.warn(
                "Storage for object of type '{}' appears to have succeeded, but this "
                "type is not officially supported!".format(type(value)))


def _h5get(store, grp, key, path=None):
    """Retrieve the underlying data for a key from its h5py container."""
    path = path + '/' + key if path else key
    result = grp[key]

    if _group_is_pandas_type(result):
        _load_pandas()
        _requires_tables()
        grp.file.flush()
        # The store must be closed for pandas to open it safely, but first we
        # copy the filename since it is not accessible after closing the file.
        # The pandas data is returned by copy, so the HDFStore can be closed.
        # Then we re-open the store.
        filename = grp.file.filename
        store.close()
        with _pandas.HDFStore(filename, mode='r') as store_:
            data = store_[path]
        store.open()
        return data
    try:
        shape = result.shape
        if shape is None:
            return None
        elif shape:
            return result
        else:
            return result[()]
    except AttributeError:
        if isinstance(result, MutableMapping):
            return H5Group(store, path)
        else:
            return result


class _ensure_open(object):

    __slots__ = ['file', 'open', 'kwargs']

    def __init__(self, file, **kwargs):
        self.file = file
        self.open = False
        self.kwargs = kwargs

    def __enter__(self):
        if self.file._file is None:
            self.file._open(** self.kwargs)
            self.open = True

    def __exit__(self, exception_type, exception_value, exception_traceback):
        if self.open:
            self.file.close()
            self.open = False


class H5Group(MutableMapping):
    """An abstraction layer over h5py's Group objects, to manage and return data."""

    __slots__ = ['_store', '_path']

    def __init__(self, store, path):
        self._store = store
        self._path = path

    def __repr__(self):
        return '{}(store={}, path={})'.format(
            type(self).__name__, repr(self._store), repr(self._path))

    @property
    def _group(self):
        return self._store.file[self._path]

    def __getitem__(self, key):
        with _ensure_open(self._store):
            return _h5get(self._store, self._group, key, self._path)

    def __setitem__(self, key, value):
        with _ensure_open(self._store):
            _h5set(self._store, self._group, self._store._validate_key(key), value, self._path)
            return value

    def __delitem__(self, key):
        with _ensure_open(self._store):
            del self._group[key]

    def __getattr__(self, name):
        with _ensure_open(self._store):
            if name in self._group.keys():
                return self.__getitem__(name)
            else:
                return getattr(self._group, name)

    def __setattr__(self, key, value):
        if key.startswith('__') or key in self.__slots__:
            super(H5Group, self).__setattr__(key, value)
        else:
            self.__setitem__(key, value)

    def setdefault(self, key, value):
        """Set a value for a key if that key is not already set."""
        super(H5Group, self).setdefault(key, value)
        return self.__getitem__(key)

    def __iter__(self):
        with _ensure_open(self._store):
            yield from self._group.keys()

    def __len__(self):
        with _ensure_open(self._store):
            return len(self._group)

    def __eq__(self, other):
        with _ensure_open(self._store):
            if isinstance(self, Mapping) and isinstance(other, Mapping):
                return super(H5Group, self).__eq__(other)
            elif type(other) == type(self):
                return self._group == other._group
            else:
                return super(H5Group, self).__eq__(other)


class H5Store(MutableMapping):
    r"""An HDF5-backed container for storing array-like and dictionary-like data.

    The H5Store is a :class:`~collections.abc.MutableMapping` and therefore
    behaves similar to a :class:`dict`, but all data is stored persistently in
    the associated HDF5 file on disk.

    Supported types include:

      * built-in types (int, float, str, bool, NoneType, array)
      * numpy arrays
      * pandas data frames (requires pandas and pytables)
      * mappings with values that are supported types

    Values can be accessed as attributes (``h5s.foo``) or via key index
    (``h5s['foo']``).

    Examples
    --------
    >>> from signac import H5Store
    >>> with H5Store('file.h5') as h5s:
    ...     h5s['foo'] = 'bar'
    ...     assert 'foo' in h5s
    ...     assert h5s.foo == 'bar'
    ...     assert h5s['foo'] == 'bar'
    >>>

    The H5Store can be used as a context manager to ensure that the underlying
    file is opened, however most built-in types (excluding arrays) can be read
    and stored without the need to _explicitly_ open the file. **To
    access arrays (reading or writing), the file must always be opened!**

    To open a file in read-only mode, use the :py:meth:`.open` method with ``mode='r'``:

    >>> with H5Store('file.h5').open(mode='r') as h5s:
    ...     pass
    >>>

    Parameters
    ----------
    filename : str
        The filename of the underlying HDF5 file.
    \*\*kwargs
        Additional keyword arguments to be forwarded to the ``h5py.File``
        constructor. See the documentation for the `h5py.File constructor
        <http://docs.h5py.org/en/latest/high/file.html#File>`_ for more
        information.

    """
    __slots__ = ['_filename', '_file', '_kwargs']

    _thread_lock = RLock()

    def __init__(self, filename, **kwargs):
        if not (isinstance(filename, str) and len(filename) > 0):
            raise ValueError('H5Store filename must be a non-empty string.')
        self._filename = os.path.realpath(filename)
        self._file = None
        self._kwargs = kwargs

    @property
    def filename(self):
        """Return the H5Store filename."""
        return self._filename

    def __repr__(self):
        return "{}(filename={})".format(type(self).__name__, repr(os.path.relpath(self._filename)))

    def __str__(self):
        return "{}(filename={})".format(type(self).__name__, repr(os.path.basename(self._filename)))

    def __del__(self):
        self.close()

    def __enter__(self):
        try:
            self.open()
        except H5StoreAlreadyOpenError:
            pass
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.close()

    def _open(self, **kwargs):
        if self._file is not None:
            raise H5StoreAlreadyOpenError(self)
        import h5py
        # We use the default file parameters, which can optionally be overridden
        # by additional keyword arguments (kwargs). This option is intentionally
        # not exposed to the public API.
        parameters = dict(self._kwargs)
        parameters.update(kwargs)
        if parameters.get('mode', None) is None:
            parameters['mode'] = 'a'

        self._thread_lock.acquire()
        try:
            self._file = h5py.File(self._filename, **parameters)
        except:  # noqa We need to release under **all** circumstances upon error!
            self._thread_lock.release()
            raise
        return self

    def open(self, mode=None):
        """Open the underlying HDF5 file.

        :param mode:
            The file open mode to use. Defaults to 'a' (append).
        :returns:
            This H5Store instance.
        """
        if mode is None:
            mode = self._kwargs.get('mode', 'a')
        return self._open(mode=mode)

    def close(self):
        """Close the underlying HDF5 file."""
        locked = True
        try:
            self._file.close()
            self._file = None
        except AttributeError:
            locked = False
        finally:
            if locked:
                try:
                    self._thread_lock.release()
                except RuntimeError as error:
                    if 'cannot release un-acquired lock' not in str(error):
                        raise

    @property
    def file(self):
        """Access the underlying instance of h5py.File.

        This property exposes the underlying ``h5py.File`` object enabling
        use of functions such as ``create_dataset()`` or ``requires_dataset()``.

        .. note::

            The store must be open to access this property!

        :returns:
            The ``h5py`` file-object that this store is operating on.
        :rtype:
            ``h5py.File``
        :raises H5StoreClosedError:
            When the store is closed at the time of accessing this property.
        """
        if self._file is None:
            raise H5StoreClosedError(self._filename)
        else:
            return self._file

    @property
    def mode(self):
        """Return the default opening mode of this H5Store."""
        return self._mode

    def flush(self):
        """Flush the underlying HDF5 file."""
        if self._file is None:
            raise H5StoreClosedError(self._filename)
        else:
            self._file.flush()

    def __getitem__(self, key):
        key = key if key.startswith('/') else '/' + key
        with _ensure_open(self):
            return _h5get(self, self._file, key)

    @staticmethod
    def _validate_key(key):
        """Emit a warning or raise an exception if key is invalid. Returns key."""
        if '.' in key:
            raise InvalidKeyError("Keys for the H5Store may not contain dots ('.').")
        return key

    def __setitem__(self, key, value):
        with _ensure_open(self):
            _h5set(self, self._file, self._validate_key(key), value)
            return value

    def __delitem__(self, key):
        with _ensure_open(self):
            del self._file[key]

    def __getattr__(self, name):
        try:
            return super(H5Store, self).__getattribute__(name)
        except AttributeError:
            if name.startswith('__') or name in self.__slots__:
                raise
            try:
                return self.__getitem__(name)
            except KeyError as e:
                raise AttributeError(e)

    def __setattr__(self, key, value):
        if key.startswith('__') or key in self.__slots__:
            super(H5Store, self).__setattr__(key, value)
        else:
            self.__setitem__(key, value)

    def __delattr__(self, key):
        if key.startswith('__') or key in self.__slots__:
            super(H5Store, self).__delattr__(key)
        else:
            self.__delitem__(key)

    def setdefault(self, key, value):
        """Set a value for a key if that key is not already set."""
        super(H5Store, self).setdefault(key, value)
        return self.__getitem__(key)

    def __iter__(self):
        with _ensure_open(self):
            yield from self._file.keys()

    def __len__(self):
        try:
            with _ensure_open(self, mode='r'):
                return len(self._file)
        except (OSError, IOError) as error:
            if 'errno = {}'.format(errno.ENOENT) in str(error):
                return 0     # file does not exist
            else:
                raise

    def __contains__(self, key):
        try:
            with _ensure_open(self, mode='r'):
                return key in self._file
        except (OSError, IOError) as error:
            if 'errno = {}'.format(errno.ENOENT) in str(error):
                return False     # file does not exist
            else:
                raise

    def clear(self):
        """Remove all data from this store.

        .. danger::
            All data will be removed, this action cannot be reversed!
        """
        with _ensure_open(self):
            self._file.clear()


class H5StoreManager(DictManager):
    """Helper class to manage multiple instances of :class:`~.H5Store` within a directory.

    Example (assuming that the 'stores/' directory exists):

    .. code-block:: python

        >>> stores = H5StoreManager('stores/')
        >>> stores.data
        <H5Store(filename=stores/data.h5)>
        >>> stores.data.foo = True
        >>> dict(stores.data)
        {'foo': True}

    :param prefix:
        The directory prefix shared by all stores managed by this class.
    """
    cls = H5Store  # type: ignore
    suffix = '.h5'  # type: ignore

    @staticmethod
    def _validate_key(key):
        """Emit a warning or raise an exception if key is invalid. Returns key."""
        if '.' in key:
            raise InvalidKeyError("Keys for the H5StoreManager may not contain dots ('.').")
        return key



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import subprocess
import logging
from os.path import expanduser
from deprecation import deprecated

import pymongo

from .errors import ConfigError
from ..version import __version__


PYMONGO_2 = pymongo.version_tuple[0] == 2

logger = logging.getLogger(__name__)

AUTH_NONE = 'none'
AUTH_SCRAM_SHA_1 = 'SCRAM-SHA-1'
AUTH_SSL = 'SSL'
AUTH_SSL_x509 = 'SSL-x509'

"""
THIS MODULE IS DEPRECATED!
"""


@deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
            details="The connection module is deprecated.")
def get_subject_from_certificate(fn_certificate):  # pragma no cover
    try:
        cert_txt = subprocess.check_output(
            ['openssl', 'x509', '-in', fn_certificate,
             '-inform', 'PEM', '-subject', '-nameopt', 'RFC2253']).decode()
    except subprocess.CalledProcessError:
        msg = "Unable to retrieve subject from certificate '{}'."
        raise RuntimeError(msg.format(fn_certificate))
    else:
        lines = cert_txt.split('\n')
        assert lines[0].startswith('subject=')
        return lines[0][len('subject='):].strip()


@deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
            details="The connection module is deprecated.")
def raise_unsupported_auth_mechanism(mechanism):
    msg = "Auth mechanism '{}' not supported."
    raise ValueError(msg.format(mechanism))


@deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
            details="The connection module is deprecated.")
class DBClientConnector(object):

    def __init__(self, host_config, **kwargs):
        self._config = host_config
        self._client = None
        self._kwargs = kwargs

    @property
    def client(self):
        if self._client is None:
            raise RuntimeError("Client not connected.")
        else:
            return self._client

    @property
    def host(self):
        return self._config['url']

    @property
    def config(self):
        return dict(self._config)

    def _config_get(self, key, default=None):
        return self._config.get(key, default)

    def _config_get_required(self, key):
        try:
            return self._config[key]
        except KeyError as e:
            raise ConfigError("Missing required key '{}'.".format(e))

    def _connect_pymongo3(self, host):
        logger.debug("Connecting with pymongo3.")
        forwarded_parameters = (
            'socketTimeoutMS', 'connectTimeoutMS', 'serverSelectionTimeoutMS',
            'w', 'wtimeout', 'replicaSet')
        parameters = self._kwargs
        for parameter in forwarded_parameters:
            if parameter in self._config:
                parameters[parameter] = self._config_get(parameter)

        auth_mechanism = self._config_get('auth_mechanism', AUTH_NONE)
        if auth_mechanism in (AUTH_NONE, AUTH_SCRAM_SHA_1):
            client = pymongo.MongoClient(
                host,
                read_preference=getattr(
                    pymongo.read_preferences.ReadPreference,
                    self._config_get('read_preference', 'PRIMARY')),
                ** parameters)
        else:
            raise_unsupported_auth_mechanism(auth_mechanism)
        self._client = client

    def connect(self, host=None):
        if host is None:
            host = self._config_get_required('url')
        logger.debug("Connecting to host '{host}'.".format(
            host=self._config_get_required('url')))

        if PYMONGO_2:
            raise RuntimeError("pymongo version 2.x is no longer supported.")
        else:
            self._connect_pymongo3(host)

    def authenticate(self):
        auth_mechanism = self._config_get('auth_mechanism', AUTH_NONE)
        logger.debug("Authenticating: mechanism={}".format(auth_mechanism))
        if auth_mechanism == AUTH_SCRAM_SHA_1:
            db_auth = self.client[self._config.get('db_auth', 'admin')]
            username = self._config_get_required('username')
            msg = "Authenticating user '{user}' with database '{db}'."
            logger.debug(msg.format(user=username, db=db_auth))
            db_auth.authenticate(
                username,
                self._config_get_required('password'),
                mechanism=AUTH_SCRAM_SHA_1)
        elif auth_mechanism in (AUTH_SSL, AUTH_SSL_x509):  # pragma no cover
            certificate_subject = get_subject_from_certificate(
                expanduser(self._config_get_required('ssl_certfile')))
            logger.debug("Authenticating: user={}".format(certificate_subject))
            db_external = self.client['$external']
            db_external.authenticate(
                certificate_subject, mechanism='MONGODB-X509')
        elif auth_mechanism == AUTH_NONE:
            pass
        else:
            raise_unsupported_auth_mechanism(auth_mechanism)

    def logout(self):
        auth_mechanism = self._config_get('auth_mechanism', AUTH_NONE)
        if auth_mechanism == AUTH_SCRAM_SHA_1:
            db_auth = self.client['admin']
            db_auth.logout()
        elif auth_mechanism in (AUTH_SSL, AUTH_SSL_x509):  # pragma no cover
            db_external = self.client['$external']
            db_external.logout()
        elif auth_mechanism == AUTH_NONE:
            pass
        else:
            raise_unsupported_auth_mechanism(auth_mechanism)



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""The signac Project and JobsCursor classes."""

import os
import stat
import re
import logging
import warnings
import errno
import uuid
import gzip
import time
from collections.abc import Iterable
from contextlib import contextmanager
from deprecation import deprecated
from itertools import groupby
from multiprocessing.pool import ThreadPool
from tempfile import TemporaryDirectory
from packaging import version

from ..version import __version__, SCHEMA_VERSION
from .. import syncutil
from ..core import json
from ..core.jsondict import JSONDict
from ..core.h5store import H5StoreManager
from .collection import Collection
from ..common.config import get_config, load_config, Config
from ..sync import sync_projects
from .job import Job
from .hashing import calc_id
from .indexing import SignacProjectCrawler
from .indexing import MainCrawler
from .utility import _mkdir_p, split_and_print_progress, _nested_dicts_to_dotted_keys
from .schema import ProjectSchema
from .errors import WorkspaceError
from .errors import DestinationExistsError
from .errors import JobsCorruptedError
from .errors import IncompatibleSchemaVersion

logger = logging.getLogger(__name__)

JOB_ID_REGEX = re.compile('[a-f0-9]{32}')

ACCESS_MODULE_MINIMAL = """import signac

def get_indexes(root):
    yield signac.get_project(root).index()
"""

ACCESS_MODULE_MAIN = """#!/usr/bin/env python
# -*- coding: utf-8 -*-
import signac

def get_indexes(root):
    yield signac.get_project(root).index()

if __name__ == '__main__':
    with signac.Collection.open('index.txt') as index:
        signac.export(signac.index(), index, update=True)
"""


class JobSearchIndex(object):
    """Search for specific jobs with filters.

    The JobSearchIndex allows to search for job_ids,
    that are part of an index, which match specific
    state point filters or job document filters.

    Parameters
    ----------
    index :
        A document index.
    _trust : bool
        Whether to skip document validation on insertion into the internal
        Collection (Default value = False).

    """
    def __init__(self, index, _trust=False):
        self._collection = Collection(index, _trust=_trust)

    def __len__(self):
        return len(self._collection)

    def _resolve_statepoint_filter(self, q):
        """Resolve state point based on filter.

        Parameters
        ----------
        q : dict
            Filter used for state point selection.

        Yields
        ------
        str
            Filtered state point.

        """
        for k, v in q.items():
            if k in ('$and', '$or'):
                if not isinstance(v, list) or isinstance(v, tuple):
                    raise ValueError(
                        "The argument to a logical operator must be a sequence (e.g. a list)!")
                yield k, [dict(self._resolve_statepoint_filter(i)) for i in v]
            else:
                yield 'statepoint.{}'.format(k), v

    def find_job_ids(self, filter=None, doc_filter=None):
        """Find job ids from a state point or document filter.

        Parameters
        ----------
        filter : dict
            A mapping of key-value pairs that all indexed job state points are
            compared against (Default value = None).
        doc_filter : dict
            A mapping of key-value pairs that all indexed job documents are
            compared against (Default value = None).

        Returns
        -------
        list
            List of job ids matching the provided filter(s).

        """
        if filter:
            filter = dict(self._resolve_statepoint_filter(filter))
            if doc_filter:
                filter.update(doc_filter)
        elif doc_filter:
            filter = doc_filter
        return self._collection._find(filter)


class _ProjectConfig(Config):
    """Extends the project config to make it immutable."""
    def __init__(self, *args, **kwargs):
        self._mutable = True
        super(_ProjectConfig, self).__init__(*args, **kwargs)
        self._mutable = False

    def __setitem__(self, key, value):
        if not self._mutable:
            warnings.warn("Modifying the project configuration after project "
                          "initialization is deprecated as of version 1.3 and "
                          "will be removed in version 2.0.",
                          DeprecationWarning)

            assert version.parse(__version__) < version.parse("2.0")
        return super(_ProjectConfig, self).__setitem__(key, value)


class Project(object):
    """The handle on a signac project.

    Application developers should usually not need to
    directly instantiate this class, but use
    :meth:`~signac.get_project` instead.

    Parameters
    ----------
    config :
        The project configuration to use. By default, it loads the first signac
        project configuration found while searching upward from the current
        working directory (Default value = None).
    _ignore_schema_version : bool
        (Default value = False).

    """
    Job = Job

    FN_DOCUMENT = 'signac_project_document.json'
    "The project's document filename."

    KEY_DATA = 'signac_data'
    "The project's datastore key."

    FN_STATEPOINTS = 'signac_statepoints.json'
    "The default filename to read from and write state points to."

    FN_CACHE = '.signac_sp_cache.json.gz'
    "The default filename for the state point cache file."

    _use_pandas_for_html_repr = True  # toggle use of pandas for html repr

    def __init__(self, config=None, _ignore_schema_version=False):
        if config is None:
            config = load_config()
        self._config = _ProjectConfig(config)

        # Ensure that the project id is configured.
        if self.id is None:
            raise LookupError(
                "Unable to determine project id. "
                "Please verify that '{}' is a signac project path.".format(
                    os.path.abspath(self.config.get('project_dir', os.getcwd()))))

        # Ensure that the project's data schema is supported.
        if not _ignore_schema_version:
            self._check_schema_compatibility()

        # Prepare project document
        self._fn_doc = os.path.join(self._rd, self.FN_DOCUMENT)
        self._document = None

        # Prepare project h5-stores
        self._stores = H5StoreManager(self._rd)

        # Prepare Workspace Directory
        if not os.path.isdir(self._wd):
            try:
                _mkdir_p(self._wd)
            except OSError:
                logger.error(
                    "Error occurred while trying to create "
                    "workspace directory for project {}.".format(self.id))
                raise

        # Internal caches
        self._index_cache = dict()
        self._sp_cache = dict()
        self._sp_cache_misses = 0
        self._sp_cache_warned = False
        self._sp_cache_miss_warning_threshold = self._config.get(
            'statepoint_cache_miss_warning_threshold', 500)

    def __str__(self):
        """Return the project's id."""
        return str(self.id)

    def __repr__(self):
        return "{type}.get_project({root})".format(
                   type=self.__class__.__name__,
                   root=repr(self.root_directory()))

    def _repr_html_(self):
        """Project details in HTML format for use in IPython environment.

        Returns
        -------
        str
            HTML containing project details.

        """
        return "<p>" + \
            '<strong>Project:</strong> {}<br>'.format(self.id) + \
            "<strong>Root:</strong> {}<br>".format(self.root_directory()) + \
            "<strong>Workspace:</strong> {}<br>".format(self.workspace()) + \
            "<strong>Size:</strong> {}".format(len(self)) + \
            "</p>" + \
            self.find_jobs()._repr_html_jobs()

    def __eq__(self, other):
        return repr(self) == repr(other)

    @property
    def config(self):
        """Get project's configuration.

        Returns
        -------
        :class:`~signac.contrib.project._ProjectConfig`
            Dictionary containing project's configuration.

        """
        return self._config

    @property
    def _rd(self):
        """Get project root directory.

        Returns
        -------
        str
            Path of project directory.

        """
        return self._config['project_dir']

    @property
    def _wd(self):
        """Get Project workspace directory.

        Returns
        -------
        str
            Path of workspace directory.

        """
        wd = os.path.expandvars(self._config.get('workspace_dir', 'workspace'))
        if os.path.isabs(wd):
            return wd
        else:
            return os.path.join(self._rd, wd)

    def root_directory(self):
        """Return the project's root directory.

        Returns
        -------
        str
            Path of project directory.

        """
        return self._rd

    def workspace(self):
        """Return the project's workspace directory.

        The workspace defaults to `project_root/workspace`.
        Configure this directory with the 'workspace_dir'
        attribute.
        If the specified directory is a relative path,
        the absolute path is relative from the project's
        root directory.

        .. note::
            The configuration will respect environment variables,
            such as ``$HOME``.

        Returns
        -------
        str
            Path of workspace directory.

        """
        return self._wd

    @deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
                details="Use project.id instead.")
    def get_id(self):
        """Get the project identifier.

        Returns
        -------
        str
            The project id.

        """
        return self.id

    @property
    def id(self):
        """Get the project identifier.

        Returns
        -------
        str
            The project id.

        """
        try:
            return str(self.config['project'])
        except KeyError:
            return None

    def _check_schema_compatibility(self):
        """Check whether this project's data schema is compatible with this version.

        Raises
        ------
        :class:`~signac.errors.IncompatibleSchemaVersion`
            If the schema version is incompatible.

        """
        schema_version = version.parse(SCHEMA_VERSION)
        config_schema_version = version.parse(self.config['schema_version'])
        if config_schema_version > schema_version:
            # Project config schema version is newer and therefore not supported.
            raise IncompatibleSchemaVersion(
                "The signac schema version used by this project is '{}', but signac {} "
                "only supports up to schema version '{}'. Try updating signac.".format(
                    config_schema_version, __version__, schema_version))
        elif config_schema_version < schema_version:
            raise IncompatibleSchemaVersion(
                "The signac schema version used by this project is '{}', but signac {} "
                "requires schema version '{}'. Please use '$ signac migrate' to "
                "irreversibly migrate this project's schema to the supported "
                "version.".format(
                    config_schema_version, __version__, schema_version))
        else:   # identical and therefore compatible
            logger.debug(
                "The project's schema version {} is supported.".format(
                    config_schema_version))

    def min_len_unique_id(self):
        """Determine the minimum length required for a job id to be unique.

        Returns
        -------
        int
            Minimum string length of a unique job identifier.

        """
        job_ids = list(self._find_job_ids())
        tmp = set()
        for i in range(32):
            tmp.clear()
            for _id in job_ids:
                if _id[:i] in tmp:
                    break
                else:
                    tmp.add(_id[:i])
            else:
                break
        return i

    def fn(self, filename):
        """Prepend a filename with the project's root directory path.

        Parameters
        ----------
        filename : str
            The name of the file.

        Returns
        -------
        str
            The joined path of project root directory and filename.

        """
        return os.path.join(self.root_directory(), filename)

    def isfile(self, filename):
        """Check if a filename exists in the project's root directory.

        Parameters
        ----------
        filename : str
            The name of the file.

        Returns
        -------
        bool
            True if filename exists in the project's root directory.

        """
        return os.path.isfile(self.fn(filename))

    def _reset_document(self, new_doc):
        """Reset document to new document passed.

        Parameters
        ----------
        new_doc : dict
            The new project document.

        """
        self.document.reset(new_doc)

    @property
    def document(self):
        """Get document associated with this project.

        Returns
        -------
        :class:`~signac.JSONDict`
            The project document.

        """
        if self._document is None:
            self._document = JSONDict(filename=self._fn_doc, write_concern=True)
        return self._document

    @document.setter
    def document(self, new_doc):
        """Setter method for document associated with this project.

        Parameters
        ----------
        new_doc : dict
            The new project document.

        """
        self._reset_document(new_doc)

    @property
    def doc(self):
        """Get document associated with this project.

        Alias for :meth:`~signac.Project.document`.

        Returns
        -------
        :class:`~signac.JSONDict`
            The project document.

        """
        return self.document

    @doc.setter
    def doc(self, new_doc):
        """Setter method for document associated with this project.

        Parameters
        ----------
        new_doc : dict
            The new project document.

        """
        self.document = new_doc

    @property
    def stores(self):
        """Get HDF5-stores associated with this project.

        Use this property to access an HDF5 file within the project's root
        directory using the H5Store dict-like interface.

        This is an example for accessing an HDF5 file called ``'my_data.h5'``
        within the project's root directory:

        .. code-block:: python

            project.stores['my_data']['array'] = np.random((32, 4))

        This is equivalent to:

        .. code-block:: python

            H5Store(project.fn('my_data.h5'))['array'] = np.random((32, 4))

        Both the `project.stores` and the `H5Store` itself support attribute
        access. The above example could therefore also be expressed as:

        .. code-block:: python

            project.stores.my_data.array = np.random((32, 4))

        Returns
        -------
        :class:`~signac.H5StoreManager`
            The HDF5-Store manager for this project.

        """
        return self._stores

    @property
    def data(self):
        """Get data associated with this project.

        This property should be used for large array-like data, which can't be
        stored efficiently in the project document. For examples and usage, see
        `Centralized Project Data
        <https://docs.signac.io/en/latest/projects.html#centralized-project-data>`_.

        Equivalent to:

        .. code-block:: python

            return project.stores['signac_data']

        See Also
        --------
        :class:`~signac.H5Store` : Usage examples.

        Returns
        -------
        :class:`~signac.H5Store`
            An HDF5-backed datastore.

        """
        return self.stores[self.KEY_DATA]

    @data.setter
    def data(self, new_data):
        """Setter method for data associated with this project.

        Parameters
        ----------
        new_data : :class:`~signac.H5Store`
            An HDF5-backed datastore.

        """
        self.stores[self.KEY_DATA] = new_data

    def open_job(self, statepoint=None, id=None):
        """Get a job handle associated with a state point.

        This method returns the job instance associated with
        the given state point or job id.
        Opening a job by a valid state point never fails.
        Opening a job by id requires a lookup of the state point
        from the job id, which may fail if the job was not
        previously initialized.

        Parameters
        ----------
        statepoint : dict
            The job's unique set of state point parameters (Default value = None).
        id : str
            The job id (Default value = None).

        Returns
        -------
        :class:`~signac.contrib.job.Job`
            The job instance.

        Raises
        ------
        KeyError
            If the attempt to open the job by id fails.
        LookupError
            If the attempt to open the job by an abbreviated id returns more
            than one match.

        """
        if (id is None) == (statepoint is None):
            raise ValueError(
                "You need to either provide the state point or the id.")
        if id is None:
            # second best case
            job = self.Job(project=self, statepoint=statepoint)
            if job._id not in self._sp_cache:
                self._sp_cache[job._id] = job.statepoint._as_dict()
            return job
        elif id in self._sp_cache:
            # optimal case
            return self.Job(project=self, statepoint=self._sp_cache[id], _id=id)
        else:
            # worst case (no state point and cache miss)
            if len(id) < 32:
                job_ids = self._find_job_ids()
                matches = [_id for _id in job_ids if _id.startswith(id)]
                if len(matches) == 1:
                    id = matches[0]
                elif len(matches) > 1:
                    raise LookupError(id)
            return self.Job(project=self, statepoint=self._get_statepoint(id), _id=id)

    def _job_dirs(self):
        """Generate ids of jobs in the workspace.

        Yields
        ------
        str
            Job id.

        """
        try:
            for d in os.listdir(self._wd):
                if JOB_ID_REGEX.match(d):
                    yield d
        except OSError as error:
            if error.errno == errno.ENOENT:
                if os.path.islink(self._wd):
                    raise WorkspaceError(
                        "The link '{}' pointing to the workspace is broken.".format(self._wd))
                elif not os.path.isdir(os.path.dirname(self._wd)):
                    logger.warning(
                        "The path to the workspace directory "
                        "('{}') does not exist.".format(os.path.dirname(self._wd)))
                else:
                    logger.info("The workspace directory '{}' does not exist!".format(self._wd))
            else:
                logger.error("Unable to access the workspace directory '{}'.".format(self._wd))
                raise WorkspaceError(error)

    def num_jobs(self):
        """Return the number of initialized jobs.

        Returns
        -------
        int
            Count of initialized jobs.

        """
        # We simply count the the number of valid directories and avoid building a list
        # for improved performance.
        i = 0
        for i, _ in enumerate(self._job_dirs(), 1):
            pass
        return i

    __len__ = num_jobs

    def __contains__(self, job):
        """Determine whether job is in the project's data space.

        Parameters
        ----------
        job : :class:`~signac.contrib.job.Job`
            The job to test for initialization.

        Returns
        -------
        bool
            True if the job is initialized for this project.

        """
        return os.path.exists(os.path.join(self._wd, job.id))

    @deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__)
    def build_job_search_index(self, index, _trust=False):
        """Build a job search index.

        Parameters
        ----------
        index : list
            A document index.
        _trust :
            (Default value = False).

        Returns
        -------
        :class:`~signac.contrib.project.JobSearchIndex`
            A job search index based on the provided index.

        """
        return JobSearchIndex(index=index, _trust=_trust)

    @deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
                details="Use the detect_schema() function instead.")
    def build_job_statepoint_index(self, exclude_const=False, index=None):
        """Build a state point index to identify jobs with specific parameters.

        This method generates pairs of state point keys and mappings of
        values to a set of all corresponding job ids. The pairs are ordered
        by the number of different values. Since state point keys may be
        nested, they are represented as a tuple.
        For example:

        .. code-block:: python

            >>> for i in range(4):
            ...     project.open_job({'a': i, 'b': {'c': i % 2}}).init()
            ...
            >>> for key, value in project.build_job_statepoint_index():
            ...     print(key)
            ...     pprint.pprint(value)
            ...
            ('b', 'c')
            defaultdict(<class 'set'>,
                        {0: {'3a530c13bfaf57517b4e81ecab6aec7f',
                             '4e9a45a922eae6bb5d144b36d82526e4'},
                         1: {'d49c6609da84251ab096654971115d0c',
                             '5c2658722218d48a5eb1e0ef7c26240b'}})
            ('a',)
            defaultdict(<class 'set'>,
                        {0: {'4e9a45a922eae6bb5d144b36d82526e4'},
                         1: {'d49c6609da84251ab096654971115d0c'},
                         2: {'3a530c13bfaf57517b4e81ecab6aec7f'},
                         3: {'5c2658722218d48a5eb1e0ef7c26240b'}})


        Values that are constant over the complete data space can be optionally
        ignored with the `exclude_const` argument set to True.

        Parameters
        ----------
        exclude_const : bool
            Exclude entries that are shared by all jobs
            that are part of the index (Default value = False).
        index :
            A document index.

        Yields
        ------
        tuple
            Pairs of state point keys and mappings of values to a set of all
            corresponding job ids (Default value = None).

        """
        from .schema import _build_job_statepoint_index
        if index is None:
            index = [{'_id': job._id, 'statepoint': job.sp()} for job in self]
        for x, y in _build_job_statepoint_index(exclude_const=exclude_const, index=index):
            yield tuple(x.split('.')), y

    def detect_schema(self, exclude_const=False, subset=None, index=None):
        """Detect the project's state point schema.

        Parameters
        ----------
        exclude_const : bool
            Exclude all state point keys that are shared by all jobs within this project
            (Default value = False).
        subset :
            A sequence of jobs or job ids specifying a subset over which the state point
            schema should be detected (Default value = None).
        index :
            A document index (Default value = None).

        Returns
        -------
        :class:`~signac.contrib.schema.ProjectSchema`
            The detected project schema.

        """
        from .schema import _build_job_statepoint_index
        if index is None:
            index = self.index(include_job_document=False)
        if subset is not None:
            subset = {str(s) for s in subset}
            index = [doc for doc in index if doc['_id'] in subset]
        statepoint_index = _build_job_statepoint_index(exclude_const=exclude_const, index=index)
        return ProjectSchema.detect(statepoint_index)

    @deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
                details="Use find_jobs().ids instead.")
    def find_job_ids(self, filter=None, doc_filter=None, index=None):
        """Find the job_ids of all jobs matching the filters.

        The optional filter arguments must be a Mapping of key-value
        pairs and JSON serializable.

        .. note::
            Providing a pre-calculated index may vastly increase the
            performance of this function.

        Parameters
        ----------
        filter : dict
            A mapping of key-value pairs that all
            indexed job state points are compared against (Default value = None).
        doc_filter : dict
            A mapping of key-value pairs that all
            indexed job documents are compared against (Default value = None).
        index :
             A document index. If not provided, an index will be computed
            (Default value = None).

        Returns
        -------
        The ids of all indexed jobs matching both filter(s).

        Raises
        ------
        TypeError
            If the filters are not JSON serializable.
        ValueError
            If the filters are invalid.
        RuntimeError
            If the filters are not supported by the index.

        """
        return self._find_job_ids(filter, doc_filter, index)

    def _find_job_ids(self, filter=None, doc_filter=None, index=None):
        """Find the job_ids of all jobs matching the filters.

        The optional filter arguments must be a JSON serializable mapping of
        key-value pairs.

        .. note::
            Providing a pre-calculated index may vastly increase the
            performance of this function.

        Parameters
        ----------
        filter : Mapping
            A mapping of key-value pairs that all indexed job state points are
            compared against (Default value = None).
        doc_filter :
            A mapping of key-value pairs that all indexed job documents are
            compared against (Default value = None).
        index :
            A document index. If not provided, an index will be computed
            (Default value = None).

        Returns
        -------
        The ids of all indexed jobs matching both filters.

        Raises
        ------
        TypeError
            If the filters are not JSON serializable.
        ValueError
            If the filters are invalid.
        RuntimeError
            If the filters are not supported by the index.

        """
        if filter is None and doc_filter is None and index is None:
            return list(self._job_dirs())
        if index is None:
            if doc_filter is None:
                index = self._sp_index()
            else:
                index = self.index(include_job_document=True)
            search_index = JobSearchIndex(index, _trust=True)
        else:
            search_index = JobSearchIndex(index)
        return search_index.find_job_ids(filter=filter, doc_filter=doc_filter)

    def find_jobs(self, filter=None, doc_filter=None):
        """Find all jobs in the project's workspace.

        The optional filter arguments must be a Mapping of key-value pairs and
        JSON serializable. The `filter` argument is used to search against job
        state points, whereas the `doc_filter` argument compares against job
        document keys.

        Parameters
        ----------
        filter : Mapping
            A mapping of key-value pairs that all indexed job state points are
            compared against (Default value = None).
        doc_filter : Mapping
            A mapping of key-value pairs that all indexed job documents are
            compared against (Default value = None).

        Returns
        -------
        :class:`~signac.contrib.project.JobsCursor`
            JobsCursor of jobs matching the provided filter(s).

        Raises
        ------
        TypeError
            If the filters are not JSON serializable.
        ValueError
            If the filters are invalid.
        RuntimeError
            If the filters are not supported by the index.

        """
        return JobsCursor(self, filter, doc_filter)

    def __iter__(self):
        return iter(self.find_jobs())

    def groupby(self, key=None, default=None):
        """Group jobs according to one or more state point parameters.

        This method can be called on any :class:`~signac.contrib.project.JobCursor` such as
        the one returned by :meth:`~signac.Project.find_jobs` or by iterating over a
        project.

        Examples
        --------
        .. code-block:: python

            # Group jobs by state point parameter 'a'.
            for key, group in project.groupby('a'):
                print(key, list(group))

            # Find jobs where job.sp['a'] is 1 and group them
            # by job.sp['b'] and job.sp['c'].
            for key, group in project.find_jobs({'a': 1}).groupby(('b', 'c')):
                print(key, list(group))

            # Group by job.sp['d'] and job.document['count'] using a lambda.
            for key, group in project.groupby(
                lambda job: (job.sp['d'], job.document['count'])
            ):
                print(key, list(group))

        If `key` is None, jobs are grouped by identity (by id), placing one job
        into each group.

        Parameters
        ----------
        key : str, iterable, or callable
            The state point grouping parameter(s) passed as a string,
            iterable of strings, or a callable that will be passed one
            argument, the job (Default value = None).
        default
            A default value to be used when a given state point key is not
            present. The value must be sortable and is only used if not None
            (Default value = None).

        Returns
        -------
        key : str
            Grouped key.
        group : iterable of Jobs
            Iterable of `Job` instances matching this group key.

        """
        return self.find_jobs().groupby(key, default=default)

    def groupbydoc(self, key=None, default=None):
        """Group jobs according to one or more document values.

        This method can be called on any :class:`~signac.contrib.project.JobCursor` such as
        the one returned by :meth:`~signac.Project.find_jobs` or by iterating over a
        project.

        Examples
        --------
        .. code-block:: python

            # Group jobs by document value 'a'.
            for key, group in project.groupbydoc('a'):
                print(key, list(group))

            # Find jobs where job.sp['a'] is 1 and group them
            # by job.document['b'] and job.document['c'].
            for key, group in project.find_jobs({'a': 1}).groupbydoc(('b', 'c')):
                print(key, list(group))

            # Group by whether 'd' is a field in the job.document using a lambda.
            for key, group in project.groupbydoc(lambda doc: 'd' in doc):
                print(key, list(group))

        If `key` is None, jobs are grouped by identity (by id), placing one job
        into each group.

        Parameters
        ----------
        key : str, iterable, or function
            The state point grouping parameter(s) passed as a string, iterable of strings,
            or a function that will be passed one argument, :meth:`~signac.job.Job.document`.
            (Default value = None).
        default :
            A default value to be used when a given state point key is not present (must
            be sortable).

        """
        return self.find_jobs().groupbydoc(key, default=default)

    def to_dataframe(self, *args, **kwargs):
        """Export the project metadata to a pandas DataFrame.

        The arguments to this function are forwarded to
        :meth:`~signac.contrib.project.JobsCursor.to_dataframe`.

        Parameters
        ----------
        *args :

        **kwargs :

        Returns
        -------
        :class:`~pandas.DataFrame`

        """
        return self.find_jobs().to_dataframe(*args, **kwargs)

    def read_statepoints(self, fn=None):
        """Read all state points from a file.

        See Also
        --------
        dump_statepoints : Dump the state points and associated job ids.
        write_statepoints : Dump state points to a file.

        Parameters
        ----------
        fn : str
            The filename of the file containing the state points,
            defaults to :attr:`~signac.Project.FN_STATEPOINTS`.

        Returns
        -------
        dict
            State points.

        """
        if fn is None:
            fn = self.fn(self.FN_STATEPOINTS)
        # See comment in write state points.
        with open(fn, 'r') as file:
            return json.loads(file.read())

    def dump_statepoints(self, statepoints):
        """Dump the state points and associated job ids.

        Equivalent to:

        .. code-block:: python

            {project.open_job(sp).id: sp for sp in statepoints}

        Parameters
        ----------
        statepoints : iterable
            A list of state points.

        Returns
        -------
        dict
            A mapping, where the key is the job id and the value is the
            state point.

        """
        return {calc_id(sp): sp for sp in statepoints}

    def write_statepoints(self, statepoints=None, fn=None, indent=2):
        """Dump state points to a file.

        If the file already contains state points, all new state points
        will be appended, while the old ones are preserved.

        See Also
        --------
        dump_statepoints : Dump the state points and associated job ids.

        Parameters
        ----------
        statepoints : iterable
            A list of state points, defaults to all state points which are
            defined in the workspace.
        fn : str
            The filename of the file containing the state points, defaults to
            :attr:`~signac.Project.FN_STATEPOINTS`.
        indent : int
            Specify the indentation of the JSON file (Default value = 2).

        """
        if fn is None:
            fn = self.fn(self.FN_STATEPOINTS)
        try:
            tmp = self.read_statepoints(fn=fn)
        except IOError as error:
            if not error.errno == errno.ENOENT:
                raise
            tmp = dict()
        if statepoints is None:
            job_ids = self._job_dirs()
            _cache = {_id: self._get_statepoint(_id) for _id in job_ids}
        else:
            _cache = {calc_id(sp): sp for sp in statepoints}

        tmp.update(_cache)
        logger.debug("Writing state points file with {} entries.".format(len(tmp)))
        with open(fn, 'w') as file:
            file.write(json.dumps(tmp, indent=indent))

    def _register(self, job):
        """Register the job within the local index.

        Parameters
        ----------
        job : :class:`~signac.contrib.job.Job`
            The job instance.

        """
        self._sp_cache[job._id] = job._statepoint._as_dict()

    def _get_statepoint_from_workspace(self, jobid):
        """Attempt to read the state point from the workspace.

        Parameters
        ----------
        jobid : str
            Identifier of the job.

        """
        fn_manifest = os.path.join(self._wd, jobid, self.Job.FN_MANIFEST)
        try:
            with open(fn_manifest, 'rb') as manifest:
                return json.loads(manifest.read().decode())
        except (IOError, ValueError) as error:
            if os.path.isdir(os.path.join(self._wd, jobid)):
                logger.error(
                    "Error while trying to access state "
                    "point manifest file of job '{}': '{}'.".format(jobid, error))
                raise JobsCorruptedError([jobid])
            raise KeyError(jobid)

    def _get_statepoint(self, jobid, fn=None):
        """Get the state point associated with a job id.

        The state point is retrieved from the internal cache, from
        the workspace or from a state points file.

        Parameters
        ----------
        jobid : str
            A job id to get the state point for.
        fn : str
            The filename of the file containing the state points, defaults
            to :attr:`~signac.Project.FN_STATEPOINTS`.

        Returns
        -------
        dict
            The state point corresponding to jobid.

        Raises
        ------
        KeyError
            If the state point associated with jobid could not be found.
        JobsCorruptedError
            If the state point manifest file corresponding to jobid is
            inaccessible or corrupted.

        """
        if not self._sp_cache:
            self._read_cache()
        try:
            if jobid in self._sp_cache:
                return self._sp_cache[jobid]
            else:
                self._sp_cache_misses += 1
                if not self._sp_cache_warned and\
                        self._sp_cache_misses > self._sp_cache_miss_warning_threshold:
                    logger.debug(
                        "High number of state point cache misses. Consider "
                        "to update cache with the Project.update_cache() method.")
                    self._sp_cache_warned = True
                sp = self._get_statepoint_from_workspace(jobid)
        except KeyError as error:
            try:
                sp = self.read_statepoints(fn=fn)[jobid]
            except IOError as io_error:
                if io_error.errno != errno.ENOENT:
                    raise io_error
                else:
                    raise error
        self._sp_cache[jobid] = sp
        return sp

    @deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
                details="Use open_job(id=jobid).statepoint() function instead.")
    def get_statepoint(self, jobid, fn=None):
        """Get the state point associated with a job id.

        The state point is retrieved from the internal cache, from
        the workspace or from a state points file.

        Parameters
        ----------
        jobid : str
            A job id to get the state point for.
        fn : str
            The filename of the file containing the state points, defaults
            to :attr:`~signac.Project.FN_STATEPOINTS`.

        Returns
        -------
        dict
            The state point corresponding to jobid.

        Raises
        ------
        KeyError
            If the state point associated with jobid could not be found.
        JobsCorruptedError
            If the state point manifest file corresponding to jobid is
            inaccessible or corrupted.

        """
        return self._get_statepoint(jobid=jobid, fn=fn)

    def create_linked_view(self, prefix=None, job_ids=None, index=None, path=None):
        """Create or update a persistent linked view of the selected data space.

        Similar to :meth:`~signac.Project.export_to`, this function expands the data space
        for the selected jobs, but instead of copying data will create symbolic links to the
        individual job workspace directories. This is primarily useful for browsing through
        the data space using a file-browser with human-interpretable directory paths.

        By default, the paths of the view will be based on variable state point keys as part
        of the *implicit* schema of the selected jobs that we create the view for. For example,
        creating a linked view for a data space with schema

        .. code-block:: python

            >>> print(project.detect_schema())
            {
             'foo': 'int([0, 1, 2, ..., 8, 9], 10)',
            }

        by calling ``project.create_linked_view('my_view')`` will look similar to:

        .. code-block:: bash

            my_view/foo/0/job -> workspace/b8fcc6b8f99c56509eb65568922e88b8
            my_view/foo/1/job -> workspace/b6cd26b873ae3624653c9268deff4485
            ...

        It is possible to control the paths using the ``path`` argument, which behaves in
        the exact same manner as the equivalent argument for :meth:`~signac.Project.export_to`.

        .. note::
            The behavior of this function is almost equivalent to
            ``project.export_to('my_view', copytree=os.symlink)`` with the
            major difference that view hierarchies are actually *updated*,
            meaning that invalid links are automatically removed.

        Parameters
        ----------
        prefix : str
            The path where the linked view will be created or updated (Default value = None).
        job_ids : iterable
            If None (the default), create the view for the complete data space,
            otherwise only for this iterable of job ids.
        index :
            A document index (Default value = None).
        path :
            The path (function) used to structure the linked data space (Default value = None).

        Returns
        -------
        dict
            A dictionary that maps the source directory paths to the linked
            directory paths.

        """
        if index is not None:
            warnings.warn(("The `index` argument is deprecated as of version 1.3 and will be "
                           "removed in version 2.0."), DeprecationWarning)

        from .linked_view import create_linked_view
        return create_linked_view(self, prefix, job_ids, index, path)

    @deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
                details="Use job.reset_statepoint() instead.")
    def reset_statepoint(self, job, new_statepoint):
        """Reset the state point of job.

        .. danger::

            Use this function with caution! Resetting a job's state point,
            may sometimes be necessary, but can possibly lead to incoherent
            data spaces.

        Parameters
        ----------
        job : :class:`~signac.contrib.job.Job`
            The job that should be reset to a new state point.
        new_statepoint : mapping
            The job's new state point.

        Raises
        ------
        :class:`~signac.errors.DestinationExistsError`
            If a job associated with the new state point is already initialized.
        OSError
            If the move failed due to an unknown system related error.

        """
        job.reset_statepoint(new_statepoint=new_statepoint)

    @deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
                details="Use job.update_statepoint() instead.")
    def update_statepoint(self, job, update, overwrite=False):
        """Update the state point of this job.

        .. warning::

            While appending to a job's state point is generally safe,
            modifying existing parameters may lead to data
            inconsistency. Use the overwrite argument with caution!

        Parameters
        ----------
        job : :class:`~signac.contrib.job.Job`
            The job whose state point shall be updated.
        update : mapping
            A mapping used for the state point update.
        overwrite :
            Set to true to ignore whether this update overwrites parameters,
            which are currently part of the job's state point. Use with caution!
            (Default value = False).

        Raises
        ------
        KeyError
            If the update contains keys, which are already part of the job's
            state point and overwrite is False.
        :class:`~signac.errors.DestinationExistsError`
            If a job associated with the new state point is already initialized.
        OSError
            If the move failed due to an unknown system related error.

        """
        job.update_statepoint(update=update, overwrite=overwrite)

    def clone(self, job, copytree=syncutil.copytree):
        """Clone job into this project.

        Create an identical copy of job within this project.

        Parameters
        ----------
        job : :class:`~signac.contrib.job.Job`
            The job to copy into this project.
        copytree :
             (Default value = syncutil.copytree)

        Returns
        -------
        :class:`~signac.contrib.job.Job`
            The job instance corresponding to the copied job.

        Raises
        ------
        :class:`~signac.errors.DestinationExistsError`
            In case that a job with the same id is already
            initialized within this project.

        """
        dst = self.open_job(job.statepoint())
        try:
            copytree(job.workspace(), dst.workspace())
        except OSError as error:
            if error.errno == errno.EEXIST:
                raise DestinationExistsError(dst)
            elif error.errno == errno.ENOENT:
                raise ValueError("Source job not initalized.")
            else:
                raise
        return dst

    def sync(self, other, strategy=None, exclude=None, doc_sync=None, selection=None, **kwargs):
        """Synchronize this project with the other project.

        Try to clone all jobs from the other project to this project.
        If a job is already part of this project, try to synchronize the job
        using the optionally specified strategies.

        Parameters
        ----------
        other : :class:`~signac.Project`
            The other project to synchronize this project with.
        strategy :
            A file synchronization strategy (Default value = None).
        exclude :
            Files with names matching the given pattern will be excluded
            from the synchronization (Default value = None).
        doc_sync :
            The function applied for synchronizing documents (Default value = None).
        selection :
            Only sync the given jobs (Default value = None).
        **kwargs :
            This method also accepts the same keyword arguments as the
            :meth:`~signac.sync.sync_projects` function.

        Raises
        ------
        :class:`~signac.errors.DocumentSyncConflict`
            If there are conflicting keys within the project or job documents that cannot
            be resolved with the given strategy or if there is no strategy provided.
        :class:`~signac.errors.FileSyncConflict`
            If there are differing files that cannot be resolved with the given strategy
            or if no strategy is provided.
        :class:`~signac.errors.SchemaSyncConflict`
            In case that the check_schema argument is True and the detected state point
            schema of this and the other project differ.

        """
        return sync_projects(
            source=other,
            destination=self,
            strategy=strategy,
            exclude=exclude,
            doc_sync=doc_sync,
            selection=selection,
            **kwargs)

    def export_to(self, target, path=None, copytree=None):
        """Export all jobs to a target location, such as a directory or a (compressed) archive file.

        Use this function in combination with :meth:`~signac.Project.find_jobs` to export only a
        select number of jobs, for example:

        .. code-block:: python

            project.find_jobs({'foo': 0}).export_to('foo_0.tar')

        The ``path`` argument enables users to control how exactly the exported data space is to be
        expanded. By default, the path-function will be based on the *implicit* schema of the
        exported jobs. For example, exporting jobs that all differ by a state point key *foo* with
        ``project.export_to('data/')``, the exported directory structure could look like this:

        .. code-block:: bash

            data/foo/0
            data/foo/1
            ...

        That would be equivalent to specifying ``path=lambda job: os.path.join('foo', job.sp.foo)``.

        Instead of a function, we can also provide a string, where fields for state point keys
        are automatically formatted. For example, the following two path arguments are equivalent:
        "foo/{foo}" and "foo/{job.sp.foo}".

        Any attribute of job can be used as a field here, so ``job.doc.bar``,
        ``job._id``, and ``job.ws`` can also be used as path fields.

        A special ``{{auto}}`` field allows us to expand the path automatically with state point
        keys that have not been specified explicitly. So, for example, one can provide
        ``path="foo/{foo}/{{auto}}"`` to specify that the path shall begin with ``foo/{foo}/``,
        but is then automatically expanded with all other state point key-value pairs. How
        key-value pairs are concatenated can be controlled *via* the format-specifier, so for
        example, ``path="{{auto:_}}"`` will generate a structure such as

        .. code-block:: bash

            data/foo_0
            data/foo_1
            ...

        Finally, providing ``path=False`` is equivalent to ``path="{job._id}"``.

        See Also
        --------
        :meth:`~signac.Project.import_from` : Previously exported or non-signac
            data spaces can be imported with.

        Parameters
        ----------
        target :
            A path to a directory to export to. The target can not already exist.
            Besides directories, possible targets are tar files (`.tar`), gzipped tar files
            (`.tar.gz`), zip files (`.zip`), bzip2-compressed files (`.bz2`),
            and xz-compressed files (`.xz`).
        path :
            The path (function) used to structure the exported data space.
            This argument must either be a callable which returns a path (str) as a function
            of `job`, a string where fields are replaced using the job-state point dictionary,
            or `False`, which means that we just use the job-id as path.
            Defaults to the equivalent of ``{{auto}}``.
        copytree :
            The function used for the actual copying of directory tree
            structures. Defaults to :func:`shutil.copytree`.
            Can only be used when the target is a directory.

        Returns
        -------
        dict
            A dict that maps the source directory paths, to the target
            directory paths.

        """
        return self.find_jobs().export_to(target=target, path=path, copytree=copytree)

    def import_from(self, origin=None, schema=None, sync=None, copytree=None):
        """Import the data space located at origin into this project.

        This function will walk through the data space located at origin and will try to identify
        data space paths that can be imported as a job workspace into this project.

        The ``schema`` argument expects a function that takes a path argument and returns a state
        point dictionary. A default function is used when no argument is provided.
        The default schema function will simply look for state point manifest files--usually named
        ``signac_statepoint.json``--and then import all data located within that path into the job
        workspace corresponding to the state point specified in the manifest file.

        Alternatively the schema argument may be a string, that is converted into a schema function,
        for example: Providing ``foo/{foo:int}`` as schema argument means that all directories under
        ``foo/`` will be imported and their names will be interpreted as the value for ``foo``
        within the state point.

        .. tip::

            Use ``copytree=os.replace`` or ``copytree=shutil.move`` to move dataspaces on import
            instead of copying them.

            Warning: Imports can fail due to conflicts. Moving data instead of copying may
            therefore lead to inconsistent states and users are advised to apply caution.

        See Also
        --------
        :meth:`~signac.Project.export_to` : Export the project data space.

        Parameters
        ----------
        origin :
            The path to the data space origin, which is to be imported. This may be a path to
            a directory, a zip file, or a tarball archive (Default value = None).
        schema :
            An optional schema function, which is either a string or a function that accepts a
            path as its first and only argument and returns the corresponding state point as dict.
            (Default value = None).
        sync :
            If ``True``, the project will be synchronized with the imported data space. If a
            dict of keyword arguments is provided, the arguments will be used for
            :meth:`~signac.Project.sync` (Default value = None).
        copytree :
            Specify which exact function to use for the actual copytree operation.
            Defaults to :func:`shutil.copytree`.

        Returns
        -------
        dict
            A dict that maps the source directory paths to the target directory paths.

        """
        from .import_export import import_into_project
        if sync:
            with self.temporary_project() as tmp_project:
                ret = tmp_project.import_from(origin=origin, schema=schema)
                if sync is True:
                    self.sync(other=tmp_project)
                else:
                    self.sync(other=tmp_project, **sync)
                return ret

        paths = dict(import_into_project(
            origin=origin, project=self, schema=schema, copytree=copytree))
        return paths

    def check(self):
        """Check the project's workspace for corruption.

        Raises
        ------
        JobsCorruptedError
            When one or more jobs are identified as corrupted.

        """
        corrupted = []
        logger.info("Checking workspace for corruption...")
        for job_id in self._find_job_ids():
            try:
                sp = self._get_statepoint(job_id)
                if calc_id(sp) != job_id:
                    corrupted.append(job_id)
                else:
                    self.open_job(sp).init()
            except JobsCorruptedError as error:
                corrupted.extend(error.job_ids)
        if corrupted:
            logger.error(
                "At least one job appears to be corrupted. Call Project.repair() "
                "to try to fix errors.")
            raise JobsCorruptedError(corrupted)

    def repair(self, fn_statepoints=None, index=None, job_ids=None):
        """Attempt to repair the workspace after it got corrupted.

        This method will attempt to repair lost or corrupted job state point
        manifest files using a state points file or a document index or both.

        Parameters
        ----------
        fn_statepoints : str
            The filename of the file containing the state points, defaults
            to :attr:`~signac.Project.FN_STATEPOINTS`.
        index :
            A document index (Default value = None).
        job_ids :
            An iterable of job ids that should get repaired. Defaults to all jobs.

        Raises
        ------
        JobsCorruptedError
            When one or more corrupted job could not be repaired.

        """
        if job_ids is None:
            job_ids = self._find_job_ids()

        # Load internal cache from all available external sources.
        self._read_cache()
        try:
            self._sp_cache.update(self.read_statepoints(fn=fn_statepoints))
        except IOError as error:
            if error.errno != errno.ENOENT or fn_statepoints is not None:
                raise
        if index is not None:
            for doc in index:
                self._sp_cache[doc['signac_id']] = doc['statepoint']

        corrupted = []
        for job_id in job_ids:
            try:
                # First, check if we can look up the state point.
                sp = self._get_statepoint(job_id)
                # Check if state point and id correspond.
                correct_id = calc_id(sp)
                if correct_id != job_id:
                    logger.warning(
                        "The job id of job '{}' is incorrect; "
                        "it should be '{}'.".format(job_id, correct_id))
                    invalid_wd = os.path.join(self.workspace(), job_id)
                    correct_wd = os.path.join(self.workspace(), correct_id)
                    try:
                        os.replace(invalid_wd, correct_wd)
                    except OSError as error:
                        logger.critical(
                            "Unable to fix location of job with "
                            " id '{}': '{}'.".format(job_id, error))
                        corrupted.append(job_id)
                        continue
                    else:
                        logger.info("Moved job to correct workspace.")

                job = self.open_job(sp)
            except KeyError:
                logger.critical("Unable to lookup state point for job with id '{}'.".format(job_id))
                corrupted.append(job_id)
            else:
                try:
                    # Try to reinit the job (triggers state point manifest file check).
                    job.init()
                except Exception as error:
                    logger.error(
                        "Error during initalization of job with "
                        "id '{}': '{}'.".format(job_id, error))
                    try:    # Attempt to fix the job manifest file.
                        job.init(force=True)
                    except Exception as error2:
                        logger.critical(
                            "Unable to force init job with id "
                            "'{}': '{}'.".format(job_id, error2))
                        corrupted.append(job_id)
        if corrupted:
            raise JobsCorruptedError(corrupted)

    def _sp_index(self):
        """Update and return the state point index cache.

        Returns
        -------
        dict
            Dictionary containing ids and state points in the cache.

        """
        job_ids = set(self._job_dirs())
        to_add = job_ids.difference(self._index_cache)
        to_remove = set(self._index_cache).difference(job_ids)
        for _id in to_remove:
            del self._index_cache[_id]
        for _id in to_add:
            self._index_cache[_id] = dict(statepoint=self._get_statepoint(_id), _id=_id)
        return self._index_cache.values()

    def _build_index(self, include_job_document=False):
        """Generate a basic state point index.

        Parameters
        ----------
        include_job_document :
            Whether to include the job document in the index (Default value =
            False).

        """
        wd = self.workspace() if self.Job is Job else None
        for _id in self._find_job_ids():
            doc = dict(_id=_id, statepoint=self._get_statepoint(_id))
            if include_job_document:
                if wd is None:
                    doc.update(self.open_job(id=_id).document)
                else:   # use optimized path
                    try:
                        with open(os.path.join(wd, _id, self.Job.FN_DOCUMENT), 'rb') as file:
                            doc.update(json.loads(file.read().decode()))
                    except IOError as error:
                        if error.errno != errno.ENOENT:
                            raise
            yield doc

    def _update_in_memory_cache(self):
        """Update the in-memory state point cache to reflect the workspace."""
        logger.debug("Updating in-memory cache...")
        start = time.time()
        job_ids = set(self._job_dirs())
        cached_ids = set(self._sp_cache)
        to_add = job_ids.difference(cached_ids)
        to_remove = cached_ids.difference(job_ids)
        if to_add or to_remove:
            for _id in to_remove:
                del self._sp_cache[_id]

            def _add(_id):
                self._sp_cache[_id] = self._get_statepoint_from_workspace(_id)

            to_add_chunks = split_and_print_progress(
                iterable=list(to_add),
                num_chunks=max(1, min(100, int(len(to_add) / 1000))),
                write=logger.info,
                desc="Read metadata: ")

            with ThreadPool() as pool:
                for chunk in to_add_chunks:
                    pool.map(_add, chunk)

            delta = time.time() - start
            logger.debug("Updated in-memory cache in {:.3f} seconds.".format(delta))
            return to_add, to_remove
        else:
            logger.debug("In-memory cache is up to date.")

    def _remove_persistent_cache_file(self):
        """Remove the persistent cache file (if it exists)."""
        try:
            os.remove(self.fn(self.FN_CACHE))
        except (OSError, IOError) as error:
            if error.errno != errno.ENOENT:
                raise error

    def update_cache(self):
        """Update the persistent state point cache.

        This function updates a persistent state point cache, which
        is stored in the project root directory. Most data space operations,
        including iteration and filtering or selection are expected
        to be significantly faster after calling this function, especially
        for large data spaces.

        """
        logger.info('Update cache...')
        start = time.time()
        cache = self._read_cache()
        self._update_in_memory_cache()
        if cache is None or set(cache) != set(self._sp_cache):
            fn_cache = self.fn(self.FN_CACHE)
            fn_cache_tmp = fn_cache + '~'
            try:
                with gzip.open(fn_cache_tmp, 'wb') as cachefile:
                    cachefile.write(json.dumps(self._sp_cache).encode())
            except OSError:  # clean-up
                try:
                    os.remove(fn_cache_tmp)
                except (OSError, IOError):
                    pass
                raise
            else:
                os.replace(fn_cache_tmp, fn_cache)
            delta = time.time() - start
            logger.info("Updated cache in {:.3f} seconds.".format(delta))
            return len(self._sp_cache)
        else:
            logger.info("Cache is up to date.")

    def _read_cache(self):
        """Read the persistent state point cache (if available)."""
        logger.debug("Reading cache...")
        start = time.time()
        try:
            with gzip.open(self.fn(self.FN_CACHE), 'rb') as cachefile:
                cache = json.loads(cachefile.read().decode())
            self._sp_cache.update(cache)
        except IOError as error:
            if not error.errno == errno.ENOENT:
                raise
            logger.debug("No cache file found.")
        else:
            delta = time.time() - start
            logger.debug("Read cache in {:.3f} seconds.".format(delta))
            return cache

    def index(self, formats=None, depth=0,
              skip_errors=False, include_job_document=True):
        r"""Generate an index of the project's workspace.

        This generator function indexes every file in the project's
        workspace until the specified `depth`.
        The job document if it exists, is always indexed, other
        files need to be specified with the formats argument.

        .. code-block:: python

            for doc in project.index({r'.*\.txt', 'TextFile'}):
                print(doc)

        Parameters
        ----------
        formats : str, dict
            The format definitions as a pattern string (e.g. ``r'.*\.txt'``)
            or a mapping from pattern strings to formats (e.g.
            ``'TextFile'``). If None, only the job document is indexed
            (Default value = None).
        depth : int
            Specifies the crawling depth. A value of 0 means no limit
            (Default value = 0).
        skip_errors : bool
            Skip all errors which occur during indexing. This is useful when
            trying to repair a broken workspace (Default value = False).
        include_job_document : bool
            Include the contents of job documents (Default value = True).

        Yields
        ------
        dict
            Index document.

        """
        if formats is None:
            root = self.workspace()

            def _full_doc(doc):
                """Add `signac_id` and `root` to the index document.

                Parameters
                ----------
                doc : dict
                    Index document.

                Returns
                -------
                dict
                    Modified index document.

                """
                doc['signac_id'] = doc['_id']
                doc['root'] = root
                return doc

            docs = self._build_index(include_job_document=include_job_document)
            docs = map(_full_doc, docs)
        else:
            if isinstance(formats, str):
                formats = {formats: 'File'}

            class Crawler(SignacProjectCrawler):
                pass
            for pattern, fmt in formats.items():
                Crawler.define(pattern, fmt)
            crawler = Crawler(self.root_directory())
            docs = crawler.crawl(depth=depth)
        if skip_errors:
            docs = _skip_errors(docs, logger.critical)
        for doc in docs:
            yield doc

    @deprecated(deprecated_in="1.5", removed_in="2.0", current_version=__version__,
                details="Access modules are deprecated.")
    def create_access_module(self, filename=None, main=True, master=None):
        """Create the access module for indexing.

        This method generates the access module required to make
        this project's index part of a main index.

        Parameters
        ----------
        filename : str
            The name of the access module file. Defaults to the standard name
            and should usually not be changed.
        main : bool
            If True, add directives for the compilation of a master index
            when executing the module (Default value = True).
        master : bool
            Deprecated parameter. Replaced by main.

        Returns
        -------
        str
            Access module name.

        """
        if master is not None:
            warnings.warn("The parameter master has been renamed to main.", DeprecationWarning)
            main = master

        if filename is None:
            filename = os.path.join(
                self.root_directory(),
                MainCrawler.FN_ACCESS_MODULE)
        with open(filename, 'x') as file:
            if main:
                file.write(ACCESS_MODULE_MAIN)
            else:
                file.write(ACCESS_MODULE_MINIMAL)
        if main:
            mode = os.stat(filename).st_mode | stat.S_IEXEC
            os.chmod(filename, mode)
        logger.info("Created access module file '{}'.".format(filename))
        return filename

    @contextmanager
    def temporary_project(self, name=None, dir=None):
        """Context manager for the initialization of a temporary project.

        The temporary project is by default created within the root project's
        workspace to ensure that they share the same file system. This is an example
        for how this method can be used for the import and synchronization of
        external data spaces.

        .. code-block:: python

            with project.temporary_project() as tmp_project:
                tmp_project.import_from('/data')
                project.sync(tmp_project)

        Parameters
        ----------
        name : str
            An optional name for the temporary project.
            Defaults to a unique random string.
        dir : str
            Optionally specify where the temporary project root directory is to be
            created. Defaults to the project's workspace directory.

        Returns
        -------
        :class:`~signac.Project`
            An instance of :class:`~signac.Project`.

        """
        if name is None:
            name = os.path.join(self.id, str(uuid.uuid4()))
        if dir is None:
            dir = self.workspace()
        _mkdir_p(self.workspace())  # ensure workspace exists
        with TemporaryProject(name=name, cls=type(self), dir=dir) as tmp_project:
            yield tmp_project

    @classmethod
    def init_project(cls, name, root=None, workspace=None, make_dir=True):
        """Initialize a project with the given name.

        It is safe to call this function multiple times with the same
        arguments. However, a `RuntimeError` is raised if an existing project
        configuration would conflict with the provided initialization
        parameters.

        Parameters
        ----------
        name : str
            The name of the project to initialize.
        root : str
            The root directory for the project.
            Defaults to the current working directory.
        workspace : str
            The workspace directory for the project.
            Defaults to a subdirectory ``workspace`` in the project root.
        make_dir : bool
            Create the project root directory if it does not exist yet
            (Default value = True).

        Returns
        -------
        :class:`~signac.Project`
            Initialized project, an instance of :class:`~signac.Project`.

        Raises
        ------
        RuntimeError
            If the project root path already contains a conflicting project
            configuration.

        """
        if root is None:
            root = os.getcwd()
        try:
            project = cls.get_project(root=root, search=False)
        except LookupError:
            fn_config = os.path.join(root, 'signac.rc')
            if make_dir:
                _mkdir_p(os.path.dirname(fn_config))
            config = get_config(fn_config)
            config['project'] = name
            if workspace is not None:
                config['workspace_dir'] = workspace
            config['schema_version'] = SCHEMA_VERSION
            config.write()
            project = cls.get_project(root=root)
            assert project.id == str(name)
            return project
        else:
            try:
                assert project.id == str(name)
                if workspace is not None:
                    assert os.path.realpath(workspace) \
                        == os.path.realpath(project.workspace())
                return project
            except AssertionError:
                raise RuntimeError(
                    "Failed to initialize project '{}'. Path '{}' already "
                    "contains a conflicting project configuration.".format(
                        name, os.path.abspath(root)))

    @classmethod
    def get_project(cls, root=None, search=True, **kwargs):
        r"""Find a project configuration and return the associated project.

        Parameters
        ----------
        root : str
            The starting point to search for a project, defaults to the
            current working directory.
        search : bool
            If True, search for project configurations inside and above
            the specified root directory, otherwise only return projects
            with a root directory identical to the specified root argument
            (Default value = True).
        \*\*kwargs :
            Optional keyword arguments that are forwarded to the
            :class:`.Project` class constructor.

        Returns
        -------
        :class:`~signac.Project`
            An instance of :class:`~signac.Project`.

        Raises
        ------
        LookupError
            When project configuration cannot be found.

        """
        if root is None:
            root = os.getcwd()
        config = load_config(root=root, local=False)
        if 'project' not in config or \
                (not search and os.path.realpath(config['project_dir']) != os.path.realpath(root)):
            raise LookupError(
                "Unable to determine project id for path '{}'.".format(os.path.abspath(root)))

        return cls(config=config, **kwargs)

    @classmethod
    def get_job(cls, root=None):
        """Find a Job in or above the current working directory (or provided path).

        Parameters
        ----------
        root : str
            The job root directory.
            If no root directory is given, the current working directory is
            assumed to be the job directory (Default value = None).

        Returns
        -------
        :class:`~signac.contrib.job.Job`
            The job instance.

        Raises
        ------
        LookupError
            When job cannot be found.

        """
        if root is None:
            root = os.getcwd()
        root = os.path.abspath(root)

        # Ensure the root path exists, which is not guaranteed by the regex match
        if not os.path.exists(root):
            raise LookupError("Path does not exist: '{}'.".format(root))

        # Find the last match instance of a job id
        results = list(re.finditer(JOB_ID_REGEX, root))
        if len(results) == 0:
            raise LookupError("Could not find a job id in path '{}'.".format(root))
        match = results[-1]
        job_id = match.group(0)
        job_root = root[:match.end()]

        # Find a project *above* the root directory (avoid finding nested projects)
        project = cls.get_project(os.path.join(job_root, os.pardir))

        # Return the matched job id from the found project
        return project.open_job(id=job_id)


@contextmanager
def TemporaryProject(name=None, cls=None, **kwargs):
    r"""Context manager for the generation of a temporary project.

    This is a factory function that creates a Project within a temporary directory
    and must be used as context manager, for example like this:

    .. code-block:: python

        with TemporaryProject() as tmp_project:
            tmp_project.import_from('/data')

    Parameters
    ----------
    name : str
        An optional name for the temporary project.
        Defaults to a unique random string.
    cls :
        The class of the temporary project.
        Defaults to :class:`~signac.Project`.
    \*\*kwargs :
        Optional keyword arguments that are forwarded to the TemporaryDirectory class
        constructor, which is used to create a temporary root directory.

    Yields
    ------
    :class:`~signac.Project`
        An instance of :class:`~signac.Project`.

    """
    if name is None:
        name = str(uuid.uuid4())
    if cls is None:
        cls = Project
    with TemporaryDirectory(**kwargs) as tmp_dir:
        yield cls.init_project(name=name, root=tmp_dir)


def _skip_errors(iterable, log=print):
    """Skip errors.

    Parameters
    ----------
    iterable : dict
        An iterable.
    log : callable
        The function to call when logging errors (Default value = print)

    Yields
    ------
    Elements from the iterable, with exceptions ignored.

    """
    while True:
        try:
            yield next(iterable)
        except StopIteration:
            return
        except Exception as error:
            log(error)


class _JobsCursorIterator(object):
    """Iterator for JobsCursor."""

    def __init__(self, project, ids):
        self._project = project
        self._ids = ids
        self._ids_iterator = iter(ids)

    def __next__(self):
        return self._project.open_job(id=next(self._ids_iterator))

    def __iter__(self):
        return type(self)(self._project, self._ids)


class JobsCursor(object):
    """An iterator over a search query result.

    Enables simple iteration and grouping operations.

    Parameters
    ----------
    project : :class:`~signac.Project`
        Project handle.
    filter : dict
        A mapping of key-value pairs that all indexed job state points are
        compared against (Default value = None).
    doc_filter : dict
        A mapping of key-value pairs that all indexed job documents are
        compared against (Default value = None).

    """
    _use_pandas_for_html_repr = True  # toggle use of pandas for html repr

    def __init__(self, project, filter, doc_filter):
        self._project = project
        self._filter = filter
        self._doc_filter = doc_filter

        # This private attribute allows us to implement the deprecated
        # next() method for this class.
        self._next_iter = None

    def __eq__(self, other):
        return self._project == other._project and self._filter == other._filter\
            and self._doc_filter == other._doc_filter

    def __len__(self):
        # Highly performance critical code path!!
        if self._filter or self._doc_filter:
            # We use the standard function for determining job ids if and only if
            # any of the two filter is provided.
            return len(self._project._find_job_ids(self._filter, self._doc_filter))
        else:
            # Without filter we can simply return the length of the whole project.
            return self._project.__len__()

    def __iter__(self):
        # Code duplication here for improved performance.
        return _JobsCursorIterator(
            self._project,
            self._project._find_job_ids(self._filter, self._doc_filter),
            )

    def next(self):
        """Return the next element.

        This function is deprecated, users should use iter(..).next() instead!
        .. deprecated:: 0.9.6

        """
        warnings.warn("Calling next() directly on a JobsCursor is deprecated!", DeprecationWarning)
        if self._next_iter is None:
            self._next_iter = iter(self)
        try:
            return next(self._next_iter)
        except StopIteration:
            self._next_iter = None
            raise

    def groupby(self, key=None, default=None):
        """Group jobs according to one or more state point parameters.

        This method can be called on any :class:`~signac.contrib.project.JobCursor` such as
        the one returned by :meth:`~signac.Project.find_jobs` or by iterating over a
        project.

        Examples
        --------
        .. code-block:: python

            # Group jobs by state point parameter 'a'.
            for key, group in project.groupby('a'):
                print(key, list(group))

            # Find jobs where job.sp['a'] is 1 and group them
            # by job.sp['b'] and job.sp['c'].
            for key, group in project.find_jobs({'a': 1}).groupby(('b', 'c')):
                print(key, list(group))

            # Group by job.sp['d'] and job.document['count'] using a lambda.
            for key, group in project.groupby(
                lambda job: (job.sp['d'], job.document['count'])
            ):
                print(key, list(group))

        If `key` is None, jobs are grouped by identity (by id), placing one job
        into each group.

        Parameters
        ----------
        key : str, iterable, or function
            The state point grouping parameter(s) passed as a string, iterable of strings,
            or a function that will be passed one argument, the job (Default value = None).
        default :
            A default value to be used when a given state point key is not present (must
            be sortable).

        """
        _filter = self._filter
        if isinstance(key, str):
            if default is None:
                if _filter is None:
                    _filter = {key: {"$exists": True}}
                else:
                    _filter = {'$and': [{key: {"$exists": True}}, _filter]}

                def keyfunction(job):
                    """Return job's state point value corresponding to the key.

                    Parameters
                    ----------
                    job : :class:`~signac.contrib.job.Job`
                        The job instance.

                    Returns
                    -------
                    State point value corresponding to the key.

                    """
                    return job.sp[key]
            else:
                def keyfunction(job):
                    """Return job's state point value corresponding to the key.

                    Return default if key is not present.

                    Parameters
                    ----------
                    job : :class:`~signac.contrib.job.Job`
                        The job instance.


                    Returns
                    -------
                    State point value corresponding to the key.
                    Default if key is not present.

                    """
                    return job.sp.get(key, default)
        elif isinstance(key, Iterable):
            if default is None:
                if _filter is None:
                    _filter = {k: {"$exists": True} for k in key}
                else:
                    _filter = {'$and': [{k: {"$exists": True} for k in key}, _filter]}

                def keyfunction(job):
                    """Return job's state point value corresponding to the key.

                    Parameters
                    ----------
                    job : :class:`~signac.contrib.job.Job`
                        The job instance.


                    Returns
                    -------
                    tuple
                        State point values.


                    """
                    return tuple(job.sp[k] for k in key)
            else:
                def keyfunction(job):
                    """Return job's state point value corresponding to the key.

                    Return default if key is not present.

                    Parameters
                    ----------
                    job : :class:`~signac.contrib.job.Job`
                        The job instance.


                    Returns
                    -------
                    tuple
                        State point values.

                    """
                    return tuple(job.sp.get(k, default) for k in key)
        elif key is None:
            # Must return a type that can be ordered with <, >
            def keyfunction(job):
                """Return the job's id.

                Parameters
                ----------
                job : :class:`~signac.contrib.job.Job`
                    The job instance.

                Returns
                -------
                str
                    The job's id.

                """
                return str(job)
        else:
            # Pass the job document to a callable
            keyfunction = key

        return groupby(sorted(iter(JobsCursor(self._project, _filter, self._doc_filter)),
                              key=keyfunction), key=keyfunction)

    def groupbydoc(self, key=None, default=None):
        """Group jobs according to one or more document values.

        This method can be called on any :class:`~signac.contrib.project.JobsCursor` such as
        the one returned by :meth:`~signac.Project.find_jobs` or by iterating over a
        project.

        Examples
        --------
        .. code-block:: python

            # Group jobs by document value 'a'.
            for key, group in project.groupbydoc('a'):
                print(key, list(group))

            # Find jobs where job.sp['a'] is 1 and group them
            # by job.document['b'] and job.document['c'].
            for key, group in project.find_jobs({'a': 1}).groupbydoc(('b', 'c')):
                print(key, list(group))

            # Group by whether 'd' is a field in the job.document using a lambda.
            for key, group in project.groupbydoc(lambda doc: 'd' in doc):
                print(key, list(group))

        If `key` is None, jobs are grouped by identity (by id), placing one job
        into each group.

        Parameters
        ----------
        key : str, iterable, or function
            The state point grouping parameter(s) passed as a string, iterable of strings,
            or a function that will be passed one argument, :meth:`~signac.job.Job.document`.
            (Default value = None).
        default :
            A default value to be used when a given state point key is not present (must
            be sortable).

        """
        if isinstance(key, str):
            if default is None:
                def keyfunction(job):
                    """Return job's document value corresponding to the key.

                    Parameters
                    ----------
                    job : :class:`~signac.contrib.job.Job`
                        The job instance.


                    Returns
                    -------
                    Document value corresponding to the key.

                    """
                    return job.document[key]
            else:
                def keyfunction(job):
                    """Return job's document value corresponding to the key.

                    Return default if key is not present.

                    Parameters
                    ----------
                    job : class:`~signac.contrib.job.Job`
                        The job instance.


                    Returns
                    -------
                    Document value corresponding to the key.
                    Default if key is not present.

                    """
                    return job.document.get(key, default)
        elif isinstance(key, Iterable):
            if default is None:
                def keyfunction(job):
                    """Return job's document value corresponding to the key.

                    Parameters
                    ----------
                    job : :class:`~signac.contrib.job.Job`
                        The job instance.


                    Returns
                    -------
                    tuple
                        Document values.

                    """
                    return tuple(job.document[k] for k in key)
            else:
                def keyfunction(job):
                    """Return job's document value corresponding to the key.

                    Return default if key is not present.

                    Parameters
                    ----------
                    job : :class:`~signac.contrib.job.Job`
                        The job instance.


                    Returns
                    -------
                    tuple
                        Document values.

                    """
                    return tuple(job.document.get(k, default) for k in key)
        elif key is None:
            # Must return a type that can be ordered with <, >
            def keyfunction(job):
                """Return the job's id.

                Parameters
                ----------
                job : :class:`~signac.contrib.job.Job`
                    The job instance.

                Returns
                -------
                str
                    The job's id.

                """
                return str(job)
        else:
            # Pass the job document to a callable
            def keyfunction(job):
                """Return job's document value corresponding to the key.

                Parameters
                ----------
                job : :class:`~signac.contrib.job.Job`
                    The job instance.

                Returns
                -------
                Document values.

                """
                return key(job.document)
        return groupby(sorted(iter(self), key=keyfunction), key=keyfunction)

    def export_to(self, target, path=None, copytree=None):
        """Export all jobs to a target location, such as a directory or a (zipped) archive file.

        See Also
        --------
        :meth:`~signac.Project.export_to` : For full details on how to use this function.

        Parameters
        ----------
        target : str
            A path to a directory or archive file to export to.
        path : str or callable
            The path (function) used to structure the exported data space
            (Default value = None).
        copytree : callable
            The function used for copying of directory tree structures.
            Defaults to :func:`shutil.copytree`. Can only be used when the
            target is a directory (Default value = None).

        Returns
        -------
        dict
            A dictionary that maps the source directory paths to the target
            directory paths.

        """
        from .import_export import export_jobs
        return dict(export_jobs(jobs=list(self), target=target,
                                path=path, copytree=copytree))

    def to_dataframe(self, sp_prefix='sp.', doc_prefix='doc.', usecols=None,
                     flatten=False):
        """Convert the selection of jobs to a pandas dataframe.

        This function exports the job metadata to a
        :py:class:`pandas.DataFrame`. All state point and document keys are
        prefixed by default to be able to distinguish them.

        Parameters
        ----------
        sp_prefix : str, optional
            Prefix state point keys with the given string. Defaults to "sp.".
        doc_prefix : str, optional
            Prefix document keys with the given string. Defaults to "doc.".
        usecols : list-like or callable, optional
            Used to select a subset of columns. If list-like, must contain
            strings corresponding to the column names that should be included.
            For example, ``['sp.a', 'doc.notes']``. If callable, the column
            will be included if the function called on the column name returns
            True. For example, ``lambda x: 'sp.' in x``. Defaults to ``None``,
            which uses all columns from the state point and document. Note
            that this filter is applied *after* the doc and sp prefixes are
            added to the column names.
        flatten : bool, optional
            Whether nested state points or document keys should be flattened.
            If True, ``{'a': {'b': 'c'}}`` becomes a column named ``a.b`` with
            value ``c``. If False, it becomes a column named ``a`` with value
            ``{'b': 'c'}``. Defaults to ``False``.

        Returns
        -------
        :class:`~pandas.DataFrame`
            A pandas DataFrame with all job metadata.

        """
        import pandas

        if usecols is None:
            def usecols(column):
                return True
        elif not callable(usecols):
            included_columns = set(usecols)

            def usecols(column):
                return column in included_columns

        def _flatten(d):
            return dict(_nested_dicts_to_dotted_keys(d)) if flatten else d

        def _export_sp_and_doc(job):
            """Prefix and filter state point and document keys.

            Parameters
            ----------
            job : :class:`~signac.contrib.job.Job`
                The job instance.

            Yields
            ------
            tuple
                tuple with prefixed state point or document key and values.

            """
            for key, value in _flatten(job.sp).items():
                prefixed_key = sp_prefix + key
                if usecols(prefixed_key):
                    yield prefixed_key, value
            for key, value in _flatten(job.doc).items():
                prefixed_key = doc_prefix + key
                if usecols(prefixed_key):
                    yield prefixed_key, value

        return pandas.DataFrame.from_dict(
            data={job._id: dict(_export_sp_and_doc(job)) for job in self},
            orient='index').infer_objects()

    def __repr__(self):
        return '{type}(project={project}, filter={filter}, doc_filter={doc_filter})'.format(
                   type=self.__class__.__name__,
                   project=repr(self._project),
                   filter=repr(self._filter),
                   doc_filter=repr(self._doc_filter))

    def _repr_html_jobs(self):
        """Jobs representation as HTML.

        Returns
        -------
        str
            HTML representation of jobs.

        """
        html = ''
        len_self = len(self)
        try:
            if len_self > 100:
                raise RuntimeError  # too large
            if self._use_pandas_for_html_repr:
                import pandas
            else:
                raise RuntimeError
        except ImportError:
            warnings.warn('Install pandas for a pretty representation of jobs.')
            html += '<br/><strong>{}</strong> job(s) found'.format(len_self)
        except RuntimeError:
            html += '<br/><strong>{}</strong> job(s) found'.format(len_self)
        else:
            with pandas.option_context("display.max_rows", 20):
                html += self.to_dataframe()._repr_html_()
        return html

    def _repr_html_(self):
        """Return an HTML representation of JobsCursor.

        Returns
        -------
        str
            HTML representation of jobs.

        """
        return repr(self) + self._repr_html_jobs()


def init_project(name, root=None, workspace=None, make_dir=True):
    """Initialize a project with the given name.

    It is safe to call this function multiple times with the same arguments.
    However, a `RuntimeError` is raised if an existing project configuration
    would conflict with the provided initialization parameters.

    Parameters
    ----------
    name : str
        The name of the project to initialize.
    root : str
        The root directory for the project.
        Defaults to the current working directory.
    workspace : str
        The workspace directory for the project.
        Defaults to a subdirectory ``workspace`` in the project root.
    make_dir : bool
        Create the project root directory, if it does not exist yet (Default
        value = True).

    Returns
    -------
    :class:`~signac.Project`
        The initialized project instance.

    Raises
    ------
    RuntimeError
        If the project root path already contains a conflicting project
        configuration.

    """
    return Project.init_project(name=name, root=root, workspace=workspace, make_dir=make_dir)


def get_project(root=None, search=True, **kwargs):
    r"""Find a project configuration and return the associated project.

    Parameters
    ----------
    root : str
        The starting point to search for a project, defaults to the current
        working directory.
    search : bool
        If True, search for project configurations inside and above the
        specified root directory, otherwise only return projects with a root
        directory identical to the specified root argument (Default value =
        True).
    \*\*kwargs :
        Optional keyword arguments that are forwarded to
        :meth:`~signac.Project.get_project`.

    Returns
    -------
    :class:`~signac.Project`
        An instance of :class:`~signac.Project`.

    Raises
    ------
    LookupError
        If no project configuration can be found.

    """
    return Project.get_project(root=root, search=search, **kwargs)


def get_job(root=None):
    """Find a Job in or above the current working directory (or provided path).

    Parameters
    ----------
    root : str
        The job root directory.
        If no root directory is given, the current working directory is
        assumed to be within the current job workspace directory (Default value = None).

    Returns
    -------
    :class:`~signac.contrib.job.Job`
        Job handle.

    Raises
    ------
    LookupError
        If this job cannot be found.

    Examples
    --------
    When the current directory is a job workspace directory:

    >>> signac.get_job()
    signac.contrib.job.Job(project=..., statepoint={...})

    """
    return Project.get_job(root=root)



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import logging
import warnings
import getpass

from deprecation import deprecated
from ..version import __version__
from ..core import json
from .config import load_config
from .errors import ConfigError, AuthenticationError
from .connection import DBClientConnector
from .crypt import get_crypt_context, SimpleKeyring, get_keyring

logger = logging.getLogger(__name__)


SESSION_PASSWORD_HASH_CACHE = SimpleKeyring()
SESSION_USERNAME_CACHE = dict()  # type: ignore

"""
THIS MODULE IS DEPRECATED!
"""


@deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
            details="The host module is deprecated.")
def get_default_host(config=None):
    if config is None:
        config = load_config()
    try:
        return config['General']['default_host']
    except KeyError:
        try:
            return config['hosts'].keys()[0]
        except (KeyError, IndexError):
            raise ConfigError("No hosts specified.")


def _get_host_config(hostname, config):
    try:
        return config['hosts'][hostname]
    except KeyError:
        raise ConfigError("Host '{}' not configured.".format(hostname))


@deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
            details="The host module is deprecated.")
def get_host_config(hostname=None, config=None):
    if config is None:
        config = load_config()
    if hostname is None:
        hostname = get_default_host(config)
    return _get_host_config(hostname, config)


def _host_id(hostcfg):
    return json.dumps(hostcfg, sort_keys=True)


@deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
            details="The host module is deprecated.")
def make_uri(hostcfg):
    ret = hostcfg['url']
    if ret.startswith('mongodb://'):
        ret = ret[len('mongodb://'):]
    if 'username' in hostcfg:
        assert '@' not in ret
        ret = hostcfg['username'] + '@' + ret
    return 'mongodb://' + ret


def _request_credentials(hostcfg):
    pwcfg = hostcfg.get('password_config')
    pw = getpass.getpass("Enter password for {}: ".format(make_uri(hostcfg)))
    if pwcfg and 'salt' in pwcfg and 'rounds' in pwcfg:
        logger.debug("Using password configuration for hashing.")
        return get_crypt_context().encrypt(pw, **pwcfg)
    else:
        return pw


def _get_config_credentials(hostcfg):
    return hostcfg.get('password')


def _get_keyring_credentials(hostcfg):
    pwcfg = hostcfg.get('password_config')
    kr = get_keyring()
    if kr is None:
        if pwcfg and 'keyring' in pwcfg:
            warnings.warn(
                "Password stored in keyring, but keyring is not available!")
    elif pwcfg and 'keyring' in pwcfg:
        return kr.get_password('signac', pwcfg['keyring'])
    else:
        return kr.get_password('signac', make_uri(hostcfg))


def _get_cached_credentials(hostcfg, default):
    hostcfg_id = json.dumps(hostcfg, sort_keys=True)
    if hostcfg_id in SESSION_PASSWORD_HASH_CACHE:
        logger.debug("Loading credentials from cache.")
        return SESSION_PASSWORD_HASH_CACHE[hostcfg_id]
    else:
        return SESSION_PASSWORD_HASH_CACHE.setdefault(hostcfg_id, default())


def _get_stored_credentials(hostcfg):
    def default():
        pw = _get_keyring_credentials(hostcfg)
        if pw is None:
            pw = _get_config_credentials(hostcfg)
        return pw
    return _get_cached_credentials(hostcfg, default)


def _get_credentials(hostcfg):
    def default():
        pw = _get_config_credentials(hostcfg)
        if pw is None:
            pw = _get_keyring_credentials(hostcfg)
            if pw is None:
                pw = _request_credentials(hostcfg)
        return pw
    return _get_cached_credentials(hostcfg, default)


@deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
            details="The host module is deprecated.")
def get_credentials(hostcfg, ask=True):
    if ask:
        return _get_credentials(hostcfg)
    else:
        return _get_stored_credentials(hostcfg)


def _input(prompt, default=''):
    try:
        value = input(prompt)
    except SyntaxError:
        return default
    if value:
        return value
    else:
        return default


@deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
            details="The host module is deprecated.")
def check_credentials(hostcfg):
    from pymongo.uri_parser import parse_uri
    auth_m = hostcfg.get('auth_mechanism', 'none')
    if auth_m == 'SCRAM-SHA-1':
        uri = hostcfg['url']
        if isinstance(uri, list):
            uri = ','.join(uri)
        if 'username' not in hostcfg and not parse_uri(uri)['username']:
            username = SESSION_USERNAME_CACHE.get(_host_id(hostcfg))
            if username:
                hostcfg['username'] = username
            else:
                SESSION_USERNAME_CACHE[_host_id(hostcfg)] = \
                    hostcfg['username'] = _input(
                        "Username ({}): ".format(getpass.getuser()),
                        getpass.getuser())
        if 'password' not in hostcfg and not parse_uri(uri)['password']:
            hostcfg['password'] = get_credentials(hostcfg)
    return hostcfg


@deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
            details="The host module is deprecated.")
def get_connector(hostcfg, **kwargs):
    return DBClientConnector(hostcfg, **kwargs)


@deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
            details="The host module is deprecated.")
def get_client(hostcfg, **kwargs):
    connector = get_connector(hostcfg, **kwargs)
    connector.connect()
    connector.authenticate()
    return connector.client


@deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
            details="The host module is deprecated.")
def get_database(name, hostname=None, config=None, **kwargs):
    if hostname is None:
        hostname = get_default_host(config)
    if config is None:
        config = load_config()
    hostcfg = check_credentials(get_host_config(hostname, config))
    logger.debug("Connecting with host config: {}".format(
        {k: '***' if 'password' in k else v for k, v in hostcfg.items()}))
    try:
        client = get_client(hostcfg, **kwargs)
    except Exception as error:
        if "Authentication failed" in str(error):
            raise AuthenticationError(hostname)
        raise
    return client[name]



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import base64

from deprecation import deprecated
from ..version import __version__

try:
    from passlib.context import CryptContext
except ImportError:
    def get_crypt_context():
        "This function requires passlib!"
        return None
else:
    def get_crypt_context():
        "Return the default signac crypto context."
        return CryptContext(schemes=('bcrypt', ))

try:
    import keyring
except ImportError:
    def get_keyring():
        "This function requires keyring!"
        return None
else:
    def get_keyring():
        "Return the system user keyring."
        # In some newer versions of keyring (probably >=21.2.0), no backend is
        # available, which causes problems for signac's implementation. This
        # signac feature is already deprecated so this is only enough of a fix
        # to prevent tests from failing for users with new versions of keyring
        # installed.
        try:
            kr = keyring.get_keyring()
        except RuntimeError:
            return None
        if kr.priority <= 0 or isinstance(kr, keyring.backends.fail.Keyring):
            return None
        else:
            return kr

"""
THIS MODULE IS DEPRECATED!
"""


@deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
            details="The crypt module is deprecated.")
class SimpleKeyring(object):
    """Simple in-memory keyring for caching."""

    def __init__(self):
        self._cache = dict()

    @classmethod
    def _encode(cls, msg):
        if msg is None:
            return
        return base64.b64encode(msg.encode())

    @classmethod
    def _decode(cls, msg):
        if msg is None:
            return
        return base64.b64decode(msg).decode()

    def __contains__(self, key):
        return key in self._cache

    def __set__(self, key, value):
        self._cache[key] = self._encode(self._secret, value)

    def __getitem__(self, key):
        return self._decode(self._cache.__getitem__(key))

    def setdefault(self, key, value):
        return self._decode(self._cache.setdefault(key, self._encode(value)))


@deprecated(deprecated_in="1.3", removed_in="2.0", current_version=__version__,
            details="The crypt module is deprecated.")
def parse_pwhash(pwhash):
    "Extract hash configuration from hash string."
    if get_crypt_context().identify(pwhash) == 'bcrypt':
        return dict(
            rounds=int(pwhash.split('$')[2]),
            salt=pwhash[-53:-31])



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Errors raised by signac."""

# The subpackage error modules (e.g. signac.core.errors) are used to bundle
# exceptions that are relevant beyond a single module. This top-level errors
# module is used to expose user-facing exception classes.

from .core.errors import Error

from .core.jsondict import BufferException
from .core.jsondict import BufferedFileError

from .common.errors import ConfigError
from .common.errors import AuthenticationError
from .common.errors import ExportError
from .common.errors import FetchError

from .contrib.errors import DestinationExistsError
from .contrib.errors import JobsCorruptedError
from .contrib.errors import IncompatibleSchemaVersion


class SyncConflict(Error, RuntimeError):
    """Raised when a synchronization operation fails."""
    pass


class FileSyncConflict(SyncConflict):
    """Raised when a synchronization operation fails due to a file conflict."""
    def __init__(self, filename):
        self.filename = filename
        "The filename of the file that caused the conflict."

    def __str__(self):
        return "The file with filename '{}' caused a conflict.".format(self.filename)


class DocumentSyncConflict(SyncConflict):
    """Raised when a synchronization operation fails due to a document conflict."""
    def __init__(self, keys):
        self.keys = keys
        "The keys that caused the conflict."

    def __str__(self):
        return "The following keys caused a conflict: {}".format(', '.join(self.keys))


class SchemaSyncConflict(SyncConflict):
    """Raised when a synchronization operation fails due to schema differences."""
    def __init__(self, schema_src, schema_dst):
        self.schema_src = schema_src
        self.schema_dst = schema_dst

    def __str__(self):
        return "The synchronization failed, because of a schema conflict."


class InvalidKeyError(ValueError):
    """Raised when a user uses a non-conforming key."""


class KeyTypeError(TypeError):
    """Raised when a user uses a key of invalid type."""


__all__ = [
    'Error',
    'BufferException',
    'BufferedFileError',
    'ConfigError',
    'AuthenticationError',
    'ExportError',
    'FetchError',
    'DestinationExistsError',
    'JobsCorruptedError',
    'IncompatibleSchemaVersion',
    'SyncConflict',
    'FileSyncConflict',
    'DocumentSyncConflict',
    'SchemaSyncConflict',
    'InvalidKeyError',
]



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Project Schema."""

from pprint import pformat
from collections import defaultdict as ddict
from numbers import Number
import itertools
from collections.abc import Mapping


class _Vividict(dict):
    """A dict that returns an empty _Vividict for keys that are missing.

    Useful for automatically nesting keys with ``vividict['a']['b']['c'] = 'd'``.

    """
    def __missing__(self, key):
        value = self[key] = type(self)()
        return value


def _collect_by_type(values):
    """Construct a mapping of types to a set of elements drawn from the input values.

    Parameters
    ----------
    values :
        An iterable of values.

    Returns
    -------
    defaultdict(set)
        A mapping of types to a set of input values of that type.

    """
    values_by_type = ddict(set)
    for v in values:
        values_by_type[type(v)].add(v)
    return values_by_type


def _build_job_statepoint_index(exclude_const, index):
    """Build index for job state points.

    Parameters
    ----------
    exclude_const : bool
        Excludes state point keys whose values are constant across the index.
    index :
        An iterable of state points.

    Yields
    ------
    statepoint_key : str
        State point key.
    statepoint_values : dict
        Dictionary mapping from a state point value to the set of job ids
        with that state point value.

    """
    from .collection import Collection
    from .collection import _DictPlaceholder
    from .utility import _nested_dicts_to_dotted_keys
    collection = Collection(index, _trust=True)
    for doc in collection.find():
        for key, _ in _nested_dicts_to_dotted_keys(doc):
            if key == '_id' or key.split('.')[0] != 'statepoint':
                continue
            collection.index(key, build=True)
    tmp = collection._indexes

    def strip_prefix(key): return k[len('statepoint.'):]

    def remove_dict_placeholder(x):
        """Remove _DictPlaceholder elements from a mapping.

        Parameters
        ----------
        x :
            Dictionary from which ``_DictPlaceholder`` values will be removed.

        Returns
        -------
        dict
            Dictionary with ``_DictPlaceholder`` keys removed.

        """
        return {k: v for k, v in x.items() if k is not _DictPlaceholder}

    for k in sorted(tmp, key=lambda k: (len(tmp[k]), k)):
        if exclude_const and len(tmp[k]) == 1 \
                and len(tmp[k][list(tmp[k].keys())[0]]) == len(collection):
            continue
        statepoint_key = strip_prefix(k)
        statepoint_values = remove_dict_placeholder(tmp[k])
        yield statepoint_key, statepoint_values


class ProjectSchema(object):
    """A description of a project's state point schema.

    Parameters
    ----------
    schema : dict
        Project schema.

    """
    def __init__(self, schema=None):
        if schema is None:
            schema = dict()
        self._schema = schema

    @classmethod
    def detect(cls, statepoint_index):
        """Detect Project's state point schema.

        Parameters
        ----------
        statepoint_index :
            State point index.

        Returns
        -------
        :class:`~ProjectSchema`
            The detected project schema.

        """
        return cls({k: _collect_by_type(v) for k, v in statepoint_index})

    def format(self, depth=None, precision=None, max_num_range=None):
        """Format the schema for printing.

        Parameters
        ----------
        depth : int
            A non-zero value will return a nested formatting up to the specified depth,
            defaults to 0.
        precision : int
            Round numerical values to the given precision, defaults to unlimited precision.
        max_num_range : int
            The maximum number of entries shown for a value range, defaults to 5.

        Returns
        -------
        str
            A formatted representation of the project schema.

        """
        if depth is None:
            depth = 0
        if max_num_range is None:
            max_num_range = 5

        def _fmt_value(x):
            """Convert a value to a string, rounded to a given precision.

            Parameters
            ----------
            x :
                Value to convert to string.

            Returns
            -------
            str
                Formatted value.

            """
            if precision is not None and isinstance(x, Number):
                return str(round(x, precision))
            else:
                return str(x)

        def _fmt_range(type_, values):
            """Convert sequence of values into a comma-separated string.

            Inserts an ellipsis (...) if the number of values exceeds ``max_num_range``.

            Parameters
            ----------
            type_ : type
                Type of values.
            values :
                An iterable of values.

            Returns
            -------
            str
                Comma-separated string of the input values.

            """
            try:
                sorted_values = sorted(values)
            except TypeError:
                sorted_values = sorted(values, key=repr)
            if len(values) <= max_num_range:
                values_string = ', '.join((_fmt_value(v) for v in sorted_values))
            else:
                values_string = ', '.join((_fmt_value(v)
                                           for v in sorted_values[:max_num_range - 2]))
                values_string += ', ..., '
                values_string += ', '.join((_fmt_value(v)
                                            for v in sorted_values[-2:]))
            return '{type_name}([{values_string}], {length})'.format(
                type_name=type_.__name__, values_string=values_string, length=len(values))

        def _fmt_values(values):
            """Convert values into a single string.

            Parameters
            ----------
            values :
                An iterable of values.

            Returns
            -------
            str
                Comma-separated string of the input values.

            """
            return ', '.join(_fmt_range(*v) for v in values.items())

        if depth > 0:
            schema_dict = _Vividict()
            for key, values in self._schema.items():
                keys = key.split('.')
                if len(keys) > 1:
                    x = schema_dict[keys[0]]
                    for k in keys[1:-1]:
                        x = x[k]
                    x[keys[-1]] = _fmt_values(values)
                else:
                    schema_dict[keys[0]] = _fmt_values(values)
            return pformat(schema_dict, depth=depth)
        else:
            ret = ['{']
            for key in sorted(self._schema):
                values = self._schema[key]
                if values:
                    ret.append(" '{}': '{}',".format(key, _fmt_values(values)))
            ret.append('}')
            return '\n'.join(ret)

    def __len__(self):
        return len(self._schema)

    def __str__(self):
        return self.format()

    def __repr__(self):
        return "{}(<len={}>)".format(type(self).__name__, len(self))

    def _repr_html_(self):
        import html
        output = "<strong>" + html.escape(repr(self)) + "</strong>"
        output += "<pre>" + str(self) + "</pre>"
        return output

    def __contains__(self, key_or_keys):
        if isinstance(key_or_keys, str):
            return key_or_keys in self._schema
        key_or_keys = '.'.join(key_or_keys)
        return key_or_keys in self._schema

    def __getitem__(self, key_or_keys):
        if isinstance(key_or_keys, str):
            return self._schema[key_or_keys]
        return self._schema['.'.join(key_or_keys)]

    def __iter__(self):
        return iter(self._schema)

    def keys(self):
        """Return schema keys."""
        return self._schema.keys()

    def values(self):
        """Return schema values."""
        return self._schema.values()

    def items(self):
        """Return schema items."""
        return self._schema.items()

    def __eq__(self, other):
        """Check if two schemas are the same.

        Returns
        -------
        bool
            True if both schemas have the same keys and values.

        """
        return self._schema == other._schema

    def difference(self, other, ignore_values=False):
        """Determine the difference between this and another project schema.

        Parameters
        ----------
        ignore_values : bool
            Ignore if the value (range) of a specific keys differ,
            only return missing keys (Default value = False).
        other : :class:`~ProjectSchema`
            Other project schema.

        Returns
        -------
        set
            A set of key tuples that are either missing or different in the other schema.

        """
        ret = set(self.keys()).difference(other.keys())
        if not ignore_values:
            ret.update({k for k, v in self.items() if k in other and other[k] != v})
        return ret

    def __call__(self, jobs_or_statepoints):
        """Evaluate the schema for the given state points.

        Parameters
        ----------
        jobs_or_statepoints :
            An iterable of jobs or state points.

        Returns
        -------
        :class:`~ProjectSchema`
            Schema of the project.

        """
        s = dict()
        iterators = itertools.tee(jobs_or_statepoints, len(self))
        for key, it in zip(self, iterators):
            values = []
            keys = key.split('.')
            for sp in it:
                if not isinstance(sp, Mapping):
                    sp = sp.statepoint
                v = sp[keys[0]]
                for k in keys[1:]:
                    v = v[k]
                values.append(v)
            s[key] = _collect_by_type(values)
        return ProjectSchema(s)



# Copyright (c) 2020 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import os
from setuptools import setup, find_packages


requirements = [
    # Deprecation management
    'deprecation>=2',
    # Platform-independent file locking
    'filelock~=3.0',
    # Used for version parsing and comparison
    'packaging>=15.0',
    # Progress bars
    'tqdm>=4.10.0',
]

description = "Manage large and heterogeneous data spaces on the file system."

try:
    this_path = os.path.dirname(os.path.abspath(__file__))
    fn_readme = os.path.join(this_path, 'README.md')
    with open(fn_readme) as fh:
        long_description = fh.read()
except (IOError, OSError):
    long_description = description

setup(
    name='signac',
    version='1.5.0',
    packages=find_packages(),
    zip_safe=True,
    maintainer='signac Developers',
    maintainer_email='signac-support@umich.edu',
    author='Carl Simon Adorf et al.',
    author_email='csadorf@umich.edu',
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://signac.io",
    download_url="https://pypi.org/project/signac/",
    keywords='simulation database index collaboration workflow',

    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Topic :: Database",
        "Topic :: Scientific/Engineering :: Physics",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],

    project_urls={
        'Homepage': 'https://signac.io',
        'Documentation': 'https://docs.signac.io',
        'Source Code': 'https://github.com/glotzerlab/signac',
        'Issue Tracker': 'https://github.com/glotzerlab/signac/issues',
    },

    install_requires=requirements,

    # Supported versions are determined according to NEP 29.
    # https://numpy.org/neps/nep-0029-deprecation_policy.html
    python_requires='>=3.6, <4',

    extras_require={
        'db': ['pymongo>=3.0'],
        'mpi': ['mpi4py'],
        'h5': ['h5py']
    },

    entry_points={
        'console_scripts': [
            'signac = signac.__main__:main',
        ],
    },
)



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Wrapper around json parsing library."""
import logging
from json import load, loads, JSONEncoder
from json.decoder import JSONDecodeError
from typing import Any, Dict, Optional

logger = logging.getLogger(__name__)

try:
    import numpy
    NUMPY = True
except ImportError:
    NUMPY = False


class CustomJSONEncoder(JSONEncoder):
    """Attempt to JSON-encode objects beyond the default supported types.

    This encoder will attempt to obtain a JSON-serializable representation of
    an object that is otherwise not serializable, by calling the object's
    `_as_dict()` method.
    """
    def default(self, o: Any) -> Dict[str, Any]:
        if NUMPY:
            if isinstance(o, numpy.number):
                return o.item()
            elif isinstance(o, numpy.ndarray):
                return o.tolist()
        try:
            return o._as_dict()
        except AttributeError:
            # Call the super method, which raises a TypeError if it cannot
            # encode the object.
            return super(CustomJSONEncoder, self).default(o)


def dumps(o: Any, sort_keys: bool = False, indent: Optional[int] = None) -> str:
    """Convert a JSON-compatible mapping into a string."""
    return CustomJSONEncoder(sort_keys=sort_keys, indent=indent).encode(o)


__all__ = ['loads', 'load', 'dumps', 'JSONDecodeError']



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Parses signac config files."""

import os
import stat
import logging

from .configobj import ConfigObj, ConfigObjError
from .validate import get_validator, cfg
from .errors import ConfigError

logger = logging.getLogger(__name__)

DEFAULT_FILENAME = '.signacrc'
CONFIG_FILENAMES = [DEFAULT_FILENAME, 'signac.rc']
HOME = os.path.expanduser('~')
CONFIG_PATH = [HOME]
FN_CONFIG = os.path.expanduser('~/.signacrc')


class PermissionsError(ConfigError):
    """Indicates an error in file permissions."""
    pass


def _search_local(root):
    for fn in CONFIG_FILENAMES:
        fn_ = os.path.abspath(os.path.join(root, fn))
        if os.path.isfile(fn_):
            yield fn_


def search_tree(root=None):
    """Locates signac configuration files in a directory hierarchy.

    Parameters
    ----------
    root : str
        Path to search. Uses ``os.getcwd()`` if None (Default value = None).

    """
    if root is None:
        root = os.getcwd()
    while(True):
        for fn in _search_local(root):
            yield fn
        up = os.path.abspath(os.path.join(root, '..'))
        if up == root:
            msg = "Reached filesystem root."
            logger.debug(msg)
            return
        else:
            root = up


def search_standard_dirs():
    """Locates signac configuration files in standard directories."""
    for path in CONFIG_PATH:
        for fn in _search_local(path):
            yield fn


def check_permissions(filename):
    """Verify that saved passwords are only readable by the current user."""
    st = os.stat(filename)
    if st.st_mode & stat.S_IROTH or st.st_mode & stat.S_IRGRP:
        raise PermissionsError("Permissions of configuration file '{fn}'"
                               "allow it to be read by others than the user. "
                               "Unable to read/write password.".format(
                                   fn=filename))


def fix_permissions(filename):
    """Set file permissions to be strictly user-readable and user-writable."""
    os.chmod(filename, stat.S_IRUSR | stat.S_IWUSR)


def check_and_fix_permissions(filename):
    """Verify file permissions and fix problems if needed."""
    try:
        check_permissions(filename)
    except PermissionsError as permissions_error:
        logger.debug(
            "{} Attempting to fix permissions.".format(permissions_error))
        try:
            fix_permissions(filename)
        except Exception as error:
            logger.error(
                "Failed to fix permissions with error: {}".format(error))
            raise permissions_error
        else:
            logger.debug("Fixed permissions.")


def read_config_file(filename):
    """Read a configuration file."""
    logger.debug("Reading config file '{}'.".format(filename))
    try:
        config = Config(filename, configspec=cfg.split('\n'))
    except (IOError, OSError, ConfigObjError) as error:
        msg = "Failed to read configuration file '{}':\n{}"
        raise ConfigError(msg.format(filename, error))
    verification = config.verify()
    if verification is not True:
        logger.debug("Config file '{}' may contain invalid values.".format(
            os.path.abspath(filename)))
    if config.has_password():
        check_and_fix_permissions(filename)
    return config


def get_config(infile=None, configspec=None, *args, **kwargs):
    """Get configuration from a file."""
    if configspec is None:
        configspec = cfg.split('\n')
    return Config(infile, configspec=configspec, *args, **kwargs)


def load_config(root=None, local=False):
    """Load configuration, searching upward from a root path."""
    if root is None:
        root = os.getcwd()
    config = Config(configspec=cfg.split('\n'))
    if local:
        for fn in _search_local(root):
            tmp = read_config_file(fn)
            config.merge(tmp)
            if 'project' in tmp:
                config['project_dir'] = os.path.dirname(fn)
                break
    else:
        for fn in search_standard_dirs():
            config.merge(read_config_file(fn))
        for fn in search_tree(root):
            tmp = read_config_file(fn)
            config.merge(tmp)
            if 'project' in tmp:
                config['project_dir'] = os.path.dirname(fn)
                break
    return config


class Config(ConfigObj):
    """Manages configuration for a signac project."""
    encoding = 'utf-8'

    def verify(self, validator=None, *args, **kwargs):
        """Validate the contents of this configuration."""
        if validator is None:
            validator = get_validator()
        return super(Config, self).validate(validator, *args, **kwargs)

    def has_password(self):
        """Check if this configuration contains a password."""
        def is_pw(section, key):
            assert not key.endswith('password')
        try:
            self.walk(is_pw)
            return False
        except AssertionError:
            return True

    def write(self, outfile=None, section=None):
        """Write this configuration to a file."""
        if outfile is not None:
            if self.has_password():
                check_and_fix_permissions(outfile)
        return super(Config, self).write(outfile=outfile, section=section)
